import hashlib
import math
import os
import gzip
import json
from functools import lru_cache
import random
import string
import sys
import uuid
from typing import List, Tuple, Union, Any, Optional
from pprint import pprint
import arrow
import click
import numpy as np
import base64
from jesse.constants import CANDLE_SOURCE_MAPPING
from jesse.constants import TIMEFRAME_PRIORITY
from jesse.constants import SUPPORTED_COLORS
from jesse.enums import timeframes

CACHED_CONFIG = dict()


def app_currency() -> str:
    from jesse.info import exchange_info
    from jesse.routes import router
    if router.routes[0].exchange in exchange_info and 'settlement_currency' in exchange_info[router.routes[0].exchange]:
        return exchange_info[router.routes[0].exchange]['settlement_currency']
    else:
        return quote_asset(router.routes[0].symbol)


@lru_cache
def app_mode() -> str:
    from jesse.config import config
    return config['app']['trading_mode']


def arrow_to_timestamp(arrow_time: arrow.arrow.Arrow) -> int:
    return arrow_time.int_timestamp * 1000


def base_asset(symbol: str) -> str:
    return symbol.split('-')[0]


def binary_search(arr: list, item) -> int:
    """
    performs a simple binary search on a sorted list

    :param arr: list
    :param item:

    :return: int
    """
    from bisect import bisect_left

    i = bisect_left(arr, item)
    if i != len(arr) and arr[i] == item:
        return i
    else:
        return -1


def class_iter(Class):
    return (value for variable, value in vars(Class).items() if
            not callable(getattr(Class, variable)) and not variable.startswith("__"))


def clean_orderbook_list(arr) -> List[List[float]]:
    return [[float(i[0]), float(i[1])] for i in arr]


def color(msg_text: str, msg_color: str) -> str:
    if not msg_text:
        return ''

    if msg_color in SUPPORTED_COLORS:
        return click.style(msg_text, fg=msg_color)
    if msg_color == 'gray':
        return click.style(msg_text, fg='white')
    raise ValueError('unsupported color')


def convert_number(old_max: float, old_min: float, new_max: float, new_min: float, old_value: float) -> float:
    """
    convert a number from one range (ex 40-119) to another
    range (ex 0-30) while keeping the ratio.
    """
    # validation
    if old_value > old_max or old_value < old_min:
        raise ValueError(f'old_value:{old_value} must be within the range. {old_min}-{old_max}')

    old_range = (old_max - old_min)
    new_range = (new_max - new_min)
    return (((old_value - old_min) * new_range) / old_range) + new_min


def dashless_symbol(symbol: str) -> str:
    return symbol.replace("-", "")


def dashy_symbol(symbol: str) -> str:
    if '-' in symbol:
        return symbol

    from jesse.config import config
    for s in config['app']['considering_symbols']:
        if dashless_symbol(s) == symbol:
            return s

    suffixes = [
        'UST', 'FDUSD', 'TUSD', 'EUT', 'EUR', 'GBP', 'JPY', 'MIM', 'TRY'
    ]

    for suffix in suffixes:
        if symbol.endswith(suffix):
            return f"{symbol[:-len(suffix)]}-{suffix}"

    if "USD" in symbol[-4:]:  # Only look at the last 4 letters
        idx = symbol.rfind("USD")
        return f"{symbol[:idx]}-{symbol[idx:]}"

    return f"{symbol[:3]}-{symbol[3:]}"


def underline_to_dashy_symbol(symbol: str) -> str:
    return symbol.replace('_', '-')


def dashy_to_underline(symbol: str) -> str:
    return symbol.replace('-', '_')


def get_base_asset(symbol: str) -> str:
    return symbol.split('-')[0]


def get_quote_asset(symbol: str) -> str:
    return symbol.split('-')[1]


def date_diff_in_days(date1: arrow.arrow.Arrow, date2: arrow.arrow.Arrow) -> int:
    if type(date1) is not arrow.arrow.Arrow or type(
            date2) is not arrow.arrow.Arrow:
        raise TypeError('dates must be Arrow instances')

    dif = date2 - date1

    return abs(dif.days)


def date_to_timestamp(date: str) -> int:
    """
    converts date string into timestamp. "2015-08-01" => 1438387200000

    :param date: str
    :return: int
    """
    return arrow_to_timestamp(arrow.get(date, 'YYYY-MM-DD'))


def dna_to_hp(strategy_hp, dna: str):
    # First, try to decode as Base64
    try:
        import base64
        import json
        # Try to decode the DNA as base64
        decoded_bytes = base64.b64decode(dna)
        # Try to parse as JSON
        decoded_json = json.loads(decoded_bytes.decode('utf-8'))
        # If successful, the decoded JSON should be the hyperparameters
        return decoded_json
    except (ValueError, base64.binascii.Error, json.JSONDecodeError):
        # If decoding fails, use the legacy method
        hp = {}

        for gene, h in zip(dna, strategy_hp):
            if h['type'] is int:
                decoded_gene = int(
                    round(
                        convert_number(119, 40, h['max'], h['min'], ord(gene))
                    )
                )
            elif h['type'] is float:
                decoded_gene = convert_number(119, 40, h['max'], h['min'], ord(gene))
            else:
                raise TypeError('Only int and float types are implemented')

            hp[h['name']] = decoded_gene

        return hp


def dump_exception() -> None:
    """
    a useful debugging helper
    """
    import traceback
    print(traceback.format_exc())
    terminate_app()


def estimate_average_price(order_qty: float, order_price: float, current_qty: float,
                           current_entry_price: float) -> float:
    """Estimates the new entry price for the position.
    This is used after having a new order and updating the currently holding position.

    Arguments:
        order_qty {float} -- qty of the new order
        order_price {float} -- price of the new order
        current_qty {float} -- current(pre-calculation) qty
        current_entry_price {float} -- current(pre-calculation) entry price

    Returns:
        float -- the new/averaged entry price
    """
    return (abs(order_qty) * order_price + abs(current_qty) *
            current_entry_price) / (abs(order_qty) + abs(current_qty))


def estimate_PNL(qty: float, entry_price: float, exit_price: float, trade_type: str, trading_fee: float = 0) -> float:
    qty = abs(qty)
    profit = qty * (exit_price - entry_price)

    if trade_type == 'short':
        profit *= -1

    fee = trading_fee * qty * (entry_price + exit_price)

    return profit - fee


def estimate_PNL_percentage(qty: float, entry_price: float, exit_price: float, trade_type: str) -> float:
    qty = abs(qty)
    profit = qty * (exit_price - entry_price)

    if trade_type == 'short':
        profit *= -1

    return (profit / (qty * entry_price)) * 100


def file_exists(path: str) -> bool:
    return os.path.isfile(path)


def clear_file(path: str) -> None:
    with open(path, 'w') as f:
        f.write('')


def make_directory(path: str) -> None:
    if not os.path.exists(path):
        os.makedirs(path)


def floor_with_precision(num: float, precision: int = 0) -> float:
    temp = 10 ** precision
    return math.floor(num * temp) / temp


def format_currency(num: float) -> str:
    return f'{num:,}'


def format_price(price: float) -> str:
    """
    Formats the price for logging.
    - If abs(price) is >= 1, it will be truncated to 2 decimal places.
    - If abs(price) is < 1, it will be truncated to 2 significant digits.
    """
    if price is None:
        return ""

    if price == 0:
        return "0.00"

    # to handle scientific notation
    price_str = f'{price:.20f}'

    if abs(price) >= 1:
        if '.' not in price_str:
            return f"{price:.2f}"

        integer_part, decimal_part = price_str.split('.')
        return f"{integer_part}.{decimal_part[:2]}"

    # For numbers between -1 and 1

    # find sign
    sign = ''
    if price < 0:
        sign = '-'

    price_str = f'{abs(price):.20f}'

    decimal_part = price_str.split('.')[1]

    first_non_zero_index = -1
    for i, digit in enumerate(decimal_part):
        if digit != '0':
            first_non_zero_index = i
            break

    # This case should ideally not be hit if price is not 0, but as a safeguard:
    if first_non_zero_index == -1:
        return "0.00"

    end_index = first_non_zero_index + 2

    formatted_decimal = decimal_part[:end_index]

    return f"{sign}0.{formatted_decimal}"


def generate_unique_id() -> str:
    return str(uuid.uuid4())


def generate_short_unique_id() -> str:
    return str(uuid.uuid4())[:22]


def get_arrow(timestamp: int) -> arrow.arrow.Arrow:
    return timestamp_to_arrow(timestamp)


def get_candle_source(candles: np.ndarray, source_type: str = "close") -> np.ndarray:
    """
    Returns the candles corresponding to the selected type.
    """
    try:
        return CANDLE_SOURCE_MAPPING[source_type](candles)
    except KeyError:
        raise ValueError(f"Source type '{source_type}' not recognised")


def get_config(keys: str, default: Any = None) -> Any:
    """
    Gets keys as a single string separated with "." and returns value.
    Also accepts a default value so that the app would work even if
    the required config value is missing from config.py file.
    Example: get_config('env.logging.order_submission', True)

    :param keys: str
    :param default: None
    :return:
    """
    if not str:
        raise ValueError('keys string cannot be empty')

    if is_unit_testing() or keys not in CACHED_CONFIG:
        if os.environ.get(keys.upper().replace(".", "_").replace(" ", "_")) is not None:
            CACHED_CONFIG[keys] = os.environ.get(keys.upper().replace(".", "_").replace(" ", "_"))
        else:
            from functools import reduce
            from jesse.config import config
            CACHED_CONFIG[keys] = reduce(lambda d, k: d.get(k, default) if isinstance(d, dict) else default,
                                         keys.split("."), config)

    return CACHED_CONFIG[keys]


def get_store():
    from jesse.store import store
    return store


def get_strategy_class(strategy_name: str):
    from pydoc import locate
    import os
    import re

    if not is_unit_testing():
        strategy_class = locate(f'strategies.{strategy_name}.{strategy_name}')
        if strategy_class is None:
            # Try to find any class that inherits from Strategy in the module
            module = locate(f'strategies.{strategy_name}')
            if module:
                strategy_file = os.path.join('strategies', strategy_name, '__init__.py')
                if os.path.exists(strategy_file):
                    with open(strategy_file, 'r') as f:
                        content = f.read()

                    # Find the class definition
                    class_pattern = r'class\s+(\w+)'
                    match = re.search(class_pattern, content)
                    if match:
                        old_class_name = match.group(1)
                        if old_class_name != strategy_name:
                            # Replace the class name in the file
                            new_content = re.sub(f'class {old_class_name}', f'class {strategy_name}', content)
                            with open(strategy_file, 'w') as f:
                                f.write(new_content)

                            # Reload the module to get the updated class
                            import importlib
                            module_path = f'strategies.{strategy_name}'
                            if module_path in sys.modules:
                                del sys.modules[module_path]
                            strategy_class = locate(f'strategies.{strategy_name}.{strategy_name}')
                            return strategy_class

                for attr_name in dir(module):
                    attr = getattr(module, attr_name)
                    if isinstance(attr, type) and attr.__module__ == f'strategies.{strategy_name}':
                        # Create a new class with the correct name as fallback
                        strategy_class = type(strategy_name, (attr,), {})
                        break
        return strategy_class

    path = sys.path[0]
    # live plugin
    if path.endswith('jesse-live'):
        strategy_dir = f'tests.strategies.{strategy_name}.{strategy_name}'
    # main framework
    else:
        strategy_dir = f'jesse.strategies.{strategy_name}.{strategy_name}'

    return locate(strategy_dir)


def insecure_hash(msg: str) -> str:
    return hashlib.md5(msg.encode()).hexdigest()


def insert_list(index: int, item, arr: list) -> list:
    """
    helper to insert an item in a Python List without removing the item
    """
    if index == -1:
        return arr + [item]

    return arr[:index] + [item] + arr[index:]


def is_backtesting() -> bool:
    from jesse.config import config
    return config['app']['trading_mode'] == 'backtest'


def is_debuggable(debug_item) -> bool:
    from jesse.config import config
    try:
        return is_debugging() and config['env']['logging'][debug_item]
    except KeyError:
        return True


def is_debugging() -> bool:
    from jesse.config import config
    return config['app']['debug_mode']


def is_importing_candles() -> bool:
    from jesse.config import config
    return config['app']['trading_mode'] == 'candles'


@lru_cache
def is_live() -> bool:
    return is_livetrading() or is_paper_trading()


@lru_cache
def is_livetrading() -> bool:
    from jesse.config import config
    return config['app']['trading_mode'] == 'livetrade'


@lru_cache
def is_optimizing() -> bool:
    from jesse.config import config
    return config['app']['trading_mode'] == 'optimize'


@lru_cache
def is_paper_trading() -> bool:
    from jesse.config import config
    return config['app']['trading_mode'] == 'papertrade'


def is_unit_testing() -> bool:
    """Returns True if the code is running by running pytest or PyCharm's test runner, False otherwise."""
    # Check if the PYTEST_CURRENT_TEST environment variable is set.
    if os.environ.get("PYTEST_CURRENT_TEST"):
        return True

    # Check if the code is being executed from the pytest command-line tool.
    script_name = os.path.basename(sys.argv[0])
    if script_name in ["pytest", "py.test"]:
        return True

    # Check if the code is being executed from PyCharm's test runner.
    if os.environ.get("PYCHARM_HOSTED"):
        return True

    # Otherwise, the code is not running by running pytest or PyCharm's test runner.
    return False


def is_valid_uuid(uuid_to_test: str, version: int = 4) -> bool:
    try:
        uuid_obj = uuid.UUID(uuid_to_test, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_to_test


def key(exchange: str, symbol: str, timeframe: str = None):
    if timeframe is None:
        return f'{exchange}-{symbol}'

    return f'{exchange}-{symbol}-{timeframe}'


def max_timeframe(timeframes_list: list) -> str:

    times = set(timeframes_list)
    for tf in TIMEFRAME_PRIORITY:
        if tf in times:
            return tf
    return timeframes.MINUTE_1


def normalize(x: float, x_min: float, x_max: float) -> float:
    """
    Rescaling data to have values between 0 and 1
    """
    return (x - x_min) / (x_max - x_min)


def now(force_fresh=False) -> int:
    """
    Always returns the current time in milliseconds but rounds time in matter of seconds
    """
    return int(now_to_timestamp(force_fresh))


def now_to_timestamp(force_fresh=False) -> int:
    if not force_fresh and (not (is_live() or is_importing_candles())):
        from jesse.store import store
        return store.app.time

    return arrow.utcnow().int_timestamp * 1000


# for use with peewee
def now_to_datetime():
    return arrow.utcnow().datetime


def current_1m_candle_timestamp():
    return arrow.utcnow().floor('minute').int_timestamp * 1000


def np_ffill(arr: np.ndarray, axis: int = 0) -> np.ndarray:
    idx_shape = tuple([slice(None)] + [np.newaxis] * (len(arr.shape) - axis - 1))
    idx = np.where(~np.isnan(arr), np.arange(arr.shape[axis])[idx_shape], 0)
    np.maximum.accumulate(idx, axis=axis, out=idx)
    slc = [
        np.arange(k)[
            tuple(
                slice(None) if dim == i else np.newaxis
                for dim in range(len(arr.shape))
            )
        ]
        for i, k in enumerate(arr.shape)
    ]

    slc[axis] = idx
    return arr[tuple(slc)]


def np_shift(arr: np.ndarray, num: int, fill_value=0) -> np.ndarray:
    result = np.empty_like(arr)

    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr

    return result


@lru_cache
def opposite_side(s: str) -> str:
    from jesse.enums import sides

    if s == sides.BUY:
        return sides.SELL
    elif s == sides.SELL:
        return sides.BUY
    else:
        raise ValueError(f'{s} is not a valid input for side')


@lru_cache
def opposite_type(t: str) -> str:
    from jesse.enums import trade_types

    if t == trade_types.LONG:
        return trade_types.SHORT
    if t == trade_types.SHORT:
        return trade_types.LONG
    raise ValueError('unsupported type')


def orderbook_insertion_index_search(arr, target: int, ascending: bool = True) -> Tuple[bool, int]:
    target = target[0]
    lower = 0
    upper = len(arr)

    while lower < upper:
        x = lower + (upper - lower) // 2
        val = arr[x][0]
        if ascending:
            if target == val:
                return True, x
            elif target > val:
                if lower == x:
                    return False, lower + 1
                lower = x
            elif target < val:
                if lower == x:
                    return False, lower
                upper = x
        elif target == val:
            return True, x
        elif target < val:
            if lower == x:
                return False, lower + 1
            lower = x
        elif target > val:
            if lower == x:
                return False, lower
            upper = x


def orderbook_trim_price(p: float, ascending: bool, unit: float) -> float:
    if ascending:
        trimmed = np.ceil(p / unit) * unit
        if math.log10(unit) < 0:
            trimmed = round(trimmed, abs(int(math.log10(unit))))
        return p if trimmed == p + unit else trimmed

    trimmed = np.ceil(p / unit) * unit - unit
    if math.log10(unit) < 0:
        trimmed = round(trimmed, abs(int(math.log10(unit))))
    return p if trimmed == p - unit else trimmed


def prepare_qty(qty: float, side: str) -> float:
    if side.lower() in ('sell', 'short'):
        return -abs(qty)
    elif side.lower() in ('buy', 'long'):
        return abs(qty)
    elif side.lower() == 'close':
        return 0.0
    else:
        raise ValueError(f'{side} is not a valid input')


def python_version() -> tuple:
    return sys.version_info[:2]


def quote_asset(symbol: str) -> str:
    try:
        return symbol.split('-')[1]
    except IndexError:
        from jesse.exceptions import InvalidRoutes
        raise InvalidRoutes(f"The symbol format is incorrect. Correct example: 'BTC-USDT'. Yours is '{symbol}'")


def random_str(num_characters: int = 8) -> str:
    return ''.join(random.choice(string.ascii_letters) for _ in range(num_characters))


def readable_duration(seconds: int, granularity: int = 2) -> str:
    intervals = (
        ('weeks', 604800),  # 60 * 60 * 24 * 7
        ('days', 86400),  # 60 * 60 * 24
        ('hours', 3600),  # 60 * 60
        ('minutes', 60),
        ('seconds', 1),
    )

    result = []
    seconds = int(seconds)

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append(f"{value} {name}")
    return ', '.join(result[:granularity])


def relative_to_absolute(path: str) -> str:
    return os.path.abspath(path)


def round_or_none(x: Union[float, None], digits: int = 0) -> Optional[float]:
    """
    Rounds a number to a certain number of digits or returns None if the number is None
    """
    if x is None:
        return None
    return round(x, digits)


def round_price_for_live_mode(price, precision: int) -> Union[float, np.ndarray]:
    """
    Rounds price(s) based on exchange requirements

    :param price: float
    :param precision: int
    :return: float | nd.array
    """
    return np.round(price, precision)


def round_qty_for_live_mode(roundable_qty: float, precision: int) -> Union[float, np.ndarray]:
    """
    Rounds qty(s) based on exchange requirements

    :param roundable_qty: float | nd.array
    :param precision: int
    :return: float | nd.array
    """
    input_type = type(roundable_qty)
    # if roundable_qty is a scalar, convert to nd.array
    if not isinstance(roundable_qty, np.ndarray):
        roundable_qty = np.array([roundable_qty])

    # for qty rounding down is important to prevent InsufficenMargin
    rounded = round_decimals_down(roundable_qty, precision)

    for index, q in enumerate(rounded):
        # if the rounded value is 0, make it the minimum possible value
        if q == 0.0:
            # if the precision is bigger or equal 0, (for numbers like 2, 0.2, 0.02)
            if precision >= 0:
                rounded[index] = 1 / 10 ** precision
            else:  # for numbers like 20, 200, 2000
                raise ValueError('qty is too small')

    if input_type in [float, np.float64]:
        return float(rounded[0])
    return rounded


def round_decimals_down(number: Union[np.ndarray, float], decimals: int = 2) -> float:
    """
    Returns a value rounded down to a specific number of decimal places.
    """
    if not isinstance(decimals, int):
        raise TypeError("decimal places must be an integer")
    elif decimals == 0:
        return np.floor(number)
    elif decimals > 0:
        factor = 10 ** decimals
        return np.floor(number * factor) / factor
    elif decimals < 0:
        # for example, for decimals = -2, we want to round down to the nearest 100 if the number is 1234, we want to return 1200:
        factor = 10 ** (decimals * -1)
        return np.floor(number / factor) * factor


def is_almost_equal(a: float, b: float, tolerance: float = 1e-8) -> bool:
    """
    Compares two floating point values with a small tolerance to account for floating point precision issues.

    :param a: First float value to compare
    :param b: Second float value to compare
    :param tolerance: The tolerance level for the comparison (default: 1e-8)
    :return: bool - True if the difference between a and b is less than or equal to the tolerance
    """
    # Check if both are None, which means they're equal
    if a is None and b is None:
        return True

    # Check if only one is None, which means they're not equal
    if a is None or b is None:
        return False

    # Check for exact equality first (optimizes for common case)
    if a == b:
        return True

    # For values very close to zero, use absolute tolerance
    if abs(a) < tolerance and abs(b) < tolerance:
        return abs(a - b) <= tolerance

    # For non-zero values, use relative tolerance
    return abs((a - b) / max(abs(a), abs(b))) <= tolerance


def same_length(bigger: np.ndarray, shorter: np.ndarray) -> np.ndarray:
    return np.concatenate((np.full((bigger.shape[0] - shorter.shape[0]), np.nan), shorter))


def secure_hash(msg: str) -> str:
    return hashlib.sha256(msg.encode()).hexdigest()


def should_execute_silently() -> bool:
    return is_optimizing() or is_unit_testing()


@lru_cache
def side_to_type(s: str) -> str:
    from jesse.enums import trade_types, sides

    # make sure string is lowercase
    s = s.lower()

    if s == sides.BUY:
        return trade_types.LONG
    if s == sides.SELL:
        return trade_types.SHORT
    raise ValueError


def string_after_character(s: str, character: str) -> str:
    try:
        return s.split(character, 1)[1]
    except IndexError:
        return None


def slice_candles(candles: np.ndarray, sequential: bool) -> np.ndarray:
    warmup_candles_num = get_config('env.data.warmup_candles_num', 240)
    if not sequential and candles.shape[0] > warmup_candles_num:
        candles = candles[-warmup_candles_num:]
    return candles


def style(msg_text: str, msg_style: str) -> str:
    if msg_style is None:
        return msg_text

    if msg_style.lower() in ['bold', 'b']:
        return click.style(msg_text, bold=True)

    if msg_style.lower() in ['underline', 'u']:
        return click.style(msg_text, underline=True)

    raise ValueError('unsupported style')


def terminate_app() -> None:
    # close the database
    from jesse.services.db import database
    database.close_connection()
    # disconnect python from the OS
    os._exit(1)


def error(msg: str, force_print: bool = False) -> None:
    # send notifications if it's a live session
    if is_live():
        from jesse.services import logger
        logger.error(msg)
        if force_print:
            _print_error(msg)
    else:
        _print_error(msg)


def _print_error(msg: str) -> None:
    print('\n')
    print(color('========== critical error =========='.upper(), 'red'))
    print(color(msg, 'red'))
    print(color('====================================', 'red'))


def timestamp_to_arrow(timestamp: int) -> arrow.arrow.Arrow:
    return arrow.get(timestamp / 1000)


def timestamp_to_date(timestamp: int) -> str:
    return str(arrow.get(timestamp / 1000))[:10]


def timestamp_to_time(timestamp: int) -> str:
    return str(arrow.get(timestamp / 1000))


def timestamp_to_iso8601(timestamp: int) -> str:
    # example: 1609804800000 => '2021-01-05T00:00:00.000Z'
    return arrow.get(timestamp / 1000).isoformat()


def iso8601_to_timestamp(iso8601: str) -> int:
    # example: '2021-01-05T00:00:00.000Z' -> 1609740800000
    return int(arrow.get(iso8601, 'YYYY-MM-DDTHH:mm:ss.SSSZ').datetime.timestamp()) * 1000


def today_to_timestamp() -> int:
    """
    returns today's (beginning) timestamp

    :return: int
    """
    return arrow.utcnow().floor('day').int_timestamp * 1000


@lru_cache
def type_to_side(t: str) -> str:
    from jesse.enums import trade_types, sides

    if t == trade_types.LONG:
        return sides.BUY
    if t == trade_types.SHORT:
        return sides.SELL
    raise ValueError(f'unsupported type: "{t}". Only "long" and "short" are supported.')


def unique_list(arr) -> list:
    """
    returns a unique version of the list while keeping its order
    :param arr: list | tuple
    :return: list
    """
    seen = set()
    seen_add = seen.add
    return [x for x in arr if not (x in seen or seen_add(x))]


def closing_side(position_type: str) -> str:
    if position_type.lower() == 'long':
        return 'sell'
    elif position_type.lower() == 'short':
        return 'buy'
    else:
        raise ValueError(f'Value entered for position_type ({position_type}) is not valid')


def merge_dicts(d1: dict, d2: dict) -> dict:
    """
    Merges nested dictionaries

    :param d1: dict
    :param d2: dict
    :return: dict
    """

    def inner(dict1, dict2):
        for k in set(dict1.keys()).union(dict2.keys()):
            if k in dict1 and k in dict2:
                if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                    yield k, dict(merge_dicts(dict1[k], dict2[k]))
                else:
                    yield k, dict2[k]
            elif k in dict1:
                yield k, dict1[k]
            else:
                yield k, dict2[k]

    return dict(inner(d1, d2))


def computer_name():
    import platform
    return platform.node()


def validate_response(response):
    if response.status_code != 200:
        err_msg = f"[{response.status_code}]: {response.json()['message']}\nPlease contact us at support@jesse.trade if this is unexpected."

        if response.status_code not in [401, 403]:
            raise ConnectionError(err_msg)
        error(err_msg, force_print=True)
        terminate_app()


def get_session_id():
    from jesse.store import store
    if store.app.session_id == '':
        store.app.session_id = generate_unique_id()
    return store.app.session_id


def get_pid():
    return os.getpid()


def is_jesse_project():
    ls = os.listdir('.')
    return 'strategies' in ls and 'storage' in ls


def dd(item):
    """
    Dump and Die but pretty: used for debugging when developing Jesse
    """
    dump(item)
    terminate_app()


def dump(*item):
    """
    Dump object in pretty format: used for debugging when developing Jesse
    """
    if len(item) == 1:
        item = item[0]

    print(
        color('\n========= Debugging Value =========='.upper(), 'yellow')
    )

    pprint(item)

    print(
        color('====================================\n', 'yellow')
    )


def debug(*item):
    """
    Used for debugging when developing Jesse. Prints the item in pretty format in both
    the terminal and the log file.
    """
    if len(item) == 1:
        dump(f"==> {item[0]}")
    else:
        dump(f"==> {', '.join(str(x) for x in item)}")
    from jesse.services import logger
    if len(item) == 1:
        logger.info(f"==> {item[0]}")
    else:
        logger.info(f"==> {', '.join(str(x) for x in item)}")


def float_or_none(item):
    """
    Return the float of the value if it's not None
    """
    if item is None or item == '':
        return None
    else:
        return float(item)


def str_or_none(item, encoding='utf-8'):
    """
    Return the str of the value if it's not None
    """
    if item is None:
        return None
    else:
        # return item if it's str, if not, decode it using encoding
        if isinstance(item, str):
            return item

        if isinstance(item, np.float64):
            return str(item)

        try:
            return str(item, encoding)
        except TypeError:
            return str(item)


def cpu_cores_count():
    from multiprocessing import cpu_count
    return cpu_count()


# a function that converts name to env_name. Example: 'Testnet Binance Futures' into 'TESTNET_BINANCE_FUTURES'
def convert_to_env_name(name: str) -> str:
    return name.replace(' ', '_').upper()


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        # Jupyter notebook or qtconsole
        if shell == 'ZMQInteractiveShell':
            return True
        elif shell == 'TerminalInteractiveShell':
            # Terminal running IPython
            return False
        else:
            # Other type (?)
            return False
    except NameError:
        # Probably standard Python interpreter
        return False


def get_os() -> str:
    import platform
    if platform.system() == 'Darwin':
        return 'mac'
    elif platform.system() == 'Linux':
        return 'linux'
    elif platform.system() == 'Windows':
        return 'windows'
    else:
        raise NotImplementedError(f'Unsupported OS: "{platform.system()}"')


# a function that returns boolean whether or not the code is being executed inside a docker container
def is_docker() -> bool:
    import os
    return os.path.exists('/.dockerenv')


def clear_output():
    if is_notebook():
        from IPython.display import clear_output
        clear_output(wait=True)
    else:
        click.clear()


def get_class_name(cls):
    # if it's a string, return it
    if isinstance(cls, str):
        return cls
    # else, return the class name
    return cls.__name__


def next_candle_timestamp(candle: np.ndarray, timeframe: str) -> int:
    return candle[0] + timeframe_to_one_minutes(timeframe) * 60_000


def get_candle_start_timestamp_based_on_timeframe(timeframe: str, num_candles_to_fetch: int) -> int:
    one_min_count = timeframe_to_one_minutes(timeframe)
    finish_date = now(force_fresh=True)
    return finish_date - (num_candles_to_fetch * one_min_count * 60_000)


def is_price_near(order_price, price_to_compare, percentage_threshold=0.00015):
    """
    Check if the given order price is near the specified price.
    Default percentage_threshold is 0.015% (0.00015)
    We calculate percentage difference between the two prices rounded to 4 decimal places,
    so low-priced orders can be properly compared within 0.015% range.
    """
    return abs(1 - (order_price / price_to_compare)) <= percentage_threshold


def gzip_compress(data):
    """Compress data using gzip."""
    json_data = json.dumps(data).encode('utf-8')
    # Compress the JSON string
    return gzip.compress(json_data)


def timeframe_to_one_minutes(timeframe: str) -> int:
    from jesse.utils import timeframe_to_one_minutes
    return timeframe_to_one_minutes(timeframe)


def compressed_response(content: str) -> dict:
    """
    Helper function to handle compression for HTTP responses.
    Returns a dict with compression info and content.

    :param content: string content to potentially compress
    :return: dict with is_compressed flag and content
    """
    # check if content is large enough to warrant compression
    compressed = gzip_compress(content)
    # encode as base64 string for safe transmission
    return {
        'is_compressed': True,
        'data': base64.b64encode(compressed).decode('utf-8')
    }


def validate_cwd() -> None:
    """
    make sure we're in a Jesse project
    """
    if not is_jesse_project():
        print(
            color(
                'Current directory is not a Jesse project. You must run commands from the root of a Jesse project. Read this page for more info: https://docs.jesse.trade/docs/getting-started/#create-a-new-jesse-project',
                'red'
            )
        )
        os._exit(1)


def has_live_trade_plugin() -> bool:
    try:
        import jesse_live
    except ModuleNotFoundError:
        return False
    return True



from jesse.strategies import Strategy


class TestLiquidationInIsolatedModeForShortTrade(Strategy):
    def on_open_position(self, order):
        assert round(self.position.liquidation_price, 2) == 14.96
        assert round(self.position.bankruptcy_price, 2) == 15

    def before(self):
        if self.index == 0:
            assert self.balance == 10000
            assert self.leverage == 2
            assert self.available_margin == 10000
            assert self.leveraged_available_margin == 20000

        # the liquidation price is at $14.94 so at $15:
        if self.price == 15:
            # assert that we are liquidated by this point
            assert self.is_close
            assert self.balance == 0
            assert self.available_margin == 0

    def should_long(self) -> bool:
        return False

    def should_short(self) -> bool:
        return self.price == 10

    def go_long(self):
        pass

    def go_short(self):
        # utils.size_to_qty(self.available_margin, self.price) == 2000
        self.sell = 2000, self.price

    def should_cancel_entry(self):
        return False



from collections import namedtuple
import numpy as np
from jesse.helpers import slice_candles
import jesse_rust

Wavetrend = namedtuple('Wavetrend', ['wt1', 'wt2', 'wtCrossUp', 'wtCrossDown', 'wtOversold', 'wtOverbought', 'wtVwap'])


# Wavetrend indicator ported from:  https://www.tradingview.com/script/Msm4SjwI-VuManChu-Cipher-B-Divergences/
#                                   https://www.tradingview.com/script/2KE8wTuF-Indicator-WaveTrend-Oscillator-WT/
#
# buySignal = wtCross and wtCrossUp and wtOversold
# sellSignal = wtCross and wtCrossDown and wtOverbought
#
# See https://github.com/ysdede/lazarus3/blob/partialexit/strategies/lazarus3/__init__.py for working jesse.ai example.

def wt(candles: np.ndarray, wtchannellen: int = 9, wtaveragelen: int = 12, wtmalen: int = 3, oblevel: int = 53, oslevel: int = -53, source_type: str = "hlc3", sequential: bool = False) -> Wavetrend:
    """
    Wavetrend indicator

    :param candles: np.ndarray
    :param wtchannellen:  int - default: 9
    :param wtaveragelen: int - default: 12
    :param wtmalen: int - default: 3
    :param oblevel: int - default: 53
    :param oslevel: int - default: -53
    :param source_type: str - default: "hlc3"
    :param sequential: bool - default: False

    :return: Wavetrend
    """
    candles = slice_candles(candles, sequential)
    
    # Use the Rust implementation
    wt1, wt2, wtCrossUp, wtCrossDown, wtOversold, wtOverbought, wtVwap = jesse_rust.wt(
        candles, 
        wtchannellen, 
        wtaveragelen, 
        wtmalen, 
        float(oblevel), 
        float(oslevel), 
        source_type
    )

    if sequential:
        return Wavetrend(wt1, wt2, wtCrossUp, wtCrossDown, wtOversold, wtOverbought, wtVwap)
    else:
        return Wavetrend(wt1[-1], wt2[-1], wtCrossUp[-1], wtCrossDown[-1], wtOversold[-1], wtOverbought[-1], wtVwap[-1])



from collections import namedtuple

from .wma import wma
from .ema import ema
import numpy as np
from jesse.helpers import get_candle_source, slice_candles

HullSuit = namedtuple('HullSuit', ['s_hull', 'm_hull', 'signal'])


def hull_suit(candles: np.ndarray, mode_switch: str = 'Hma', length: int = 55, length_mult: float = 1.0, source_type: str = 'close', sequential: bool = False) -> HullSuit:
    """
    @author InSilico
    credits: https://www.tradingview.com/script/hg92pFwS-Hull-Suite/

    HullSuit - Hull Suit

    :param candles: np.ndarray
    :param mode_switch: str - default: 'Hma'
    :param length: int - default: 55
    :param length_mult: float - default: 1.0
    :param source_type: str - default: "closes"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    mode_len = int(length * length_mult)
    if mode_switch == 'Hma':
        mode = wma(2*wma(source, int(mode_len / 2), sequential=True) - wma(source, mode_len, sequential=True), round(mode_len ** 0.5), sequential=True)
    elif mode_switch == 'Ehma':
        mode = ema(2*ema(source, int(mode_len / 2), sequential=True) - ema(source,
                   mode_len, sequential=True), round(mode_len ** 0.5), sequential=True)
    elif mode_switch == 'Thma':
        mode = wma(3*wma(source, int(mode_len / 6), sequential=True) - wma(source, int(mode_len / 4), sequential=True) -
                   wma(source, int(mode_len / 2), sequential=True), int(mode_len / 2), sequential=True)

    # Vectorized computation for s_hull, m_hull, and signal
    n = len(mode)
    s_hull = np.full(n, None, dtype=object)
    m_hull = np.full(n, None, dtype=object)
    signal = np.full(n, None, dtype=object)
    if n > 2:
        s_hull[2:] = mode[:-2]
        m_hull[2:] = mode[2:]
        signal[2:] = np.where(mode[:-2] < mode[2:], 'buy', 'sell')

    if sequential:
        return HullSuit(s_hull, m_hull, signal)
    else:
        return HullSuit(s_hull[-1], m_hull[-1], signal[-1])



from typing import Union

import numpy as np

try:
    from numba import njit
except ImportError:
    njit = lambda a : a

from jesse.helpers import get_candle_source, slice_candles


def supersmoother_3_pole(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> \
        Union[
            float, np.ndarray]:
    """
    Super Smoother Filter 3pole Butterworth
    This indicator was described by John F. Ehlers

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = supersmoother_fast(source, period)

    return res if sequential else res[-1]


@njit(cache=True)
def supersmoother_fast(source, period):
    a = np.exp(-np.pi / period)
    b = 2 * a * np.cos(1.738 * np.pi / period)
    c = a ** 2
    newseries = np.copy(source)
    for i in range(3, source.shape[0]):
        newseries[i] = (1 - c ** 2 - b + b * c) * source[i] \
                       + (b + c) * newseries[i - 1] + (-c - b * c) * newseries[i - 2] + (c ** 2) * newseries[i - 3]
    return newseries



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import vwma as vwma_rust


def vwma(candles: np.ndarray, period: int = 20, source_type: str = "close", sequential: bool = False) -> Union[
        float, np.ndarray]:
    """
    VWMA - Volume Weighted Moving Average

    :param candles: np.ndarray
    :param period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        # Handle 1D array case by creating dummy candles with volume=1
        source = candles
        dummy_candles = np.column_stack([
            np.zeros(len(source)),  # timestamp
            np.zeros(len(source)),  # open
            source,                 # close
            np.zeros(len(source)),  # high
            np.zeros(len(source)),  # low
            np.ones(len(source))    # volume
        ])
        candles_f64 = np.asarray(dummy_candles, dtype=np.float64)
    else:
        candles = slice_candles(candles, sequential)
        # Convert to float64 for Rust compatibility
        candles_f64 = np.asarray(candles, dtype=np.float64)
        
        # Update close column if different source type is requested
        if source_type != "close":
            source = get_candle_source(candles, source_type=source_type)
            candles_f64[:, 2] = source

    # Call the Rust implementation
    result = vwma_rust(candles_f64, period)
    
    return result if sequential else result[-1]



import requests
import jesse.helpers as jh
from timeloop import Timeloop
from datetime import timedelta

MSG_QUEUE = []


def start_notifier_loop():
    """
    a constant running loop that runs in a separate thread and
    checks for new messages in the msg_queue. If there are
    any, it sends them by calling _telegram() and _discord()
    """
    from jesse.store import store
    from jesse.services.transformers import get_notification_api_key

    tl = Timeloop()
    @tl.job(interval=timedelta(seconds=0.5))
    def handle_time():
        if len(MSG_QUEUE) > 0:
            msg = MSG_QUEUE.pop(0)

            notification_keys = get_notification_api_key(store.app.notifications_api_key, protect_sensitive_data=False) if store.app.notifications_api_key else None

            if msg['type'] == 'info':
                if msg['webhook'] is None and notification_keys:
                    if notification_keys['driver'] == 'telegram':
                        _telegram(msg['content'], notification_keys['bot_token'], notification_keys['chat_id'])
                    elif notification_keys['driver'] == 'discord':
                        _discord(msg['content'], webhook_address=notification_keys['webhook'])
                    elif notification_keys['driver'] == 'slack':
                        _slack(msg['content'], webhook_address=notification_keys['webhook'])
                elif notification_keys:
                    _custom_channel_notification(msg)

            # elif msg['type'] == 'error' and error_notifications:
            #     if error_notifications['driver'] == 'telegram':
            #         _telegram_errors(msg['content'], error_notifications['bot_token'], error_notifications['chat_id'])
            #     elif error_notifications['driver'] == 'discord':
            #         _discord(msg['content'], webhook_address=error_notifications['webhook'])
            #     elif error_notifications['driver'] == 'slack':
            #         _slack(msg['content'], webhook_address=error_notifications['webhook'])
            else:
                raise ValueError(f'Unknown message type: {msg["type"]}')

    tl.start()


def notify(msg: str, webhook=None) -> None:
    """
    sends notifications to "main_telegram_bot" which is supposed to receive messages.
    """
    msg = _format_msg(msg)

    # Notification drivers don't accept text with more than 2000 characters.
    # So if that's the case limit it to the last 2000 characters.
    if len(msg) > 2000:
        msg = msg[-2000:]

    MSG_QUEUE.append({'type': 'info', 'content': msg, 'webhook': webhook})


def _telegram(msg: str, token: str, chat_id: str) -> None:
    from jesse.services import logger

    if not token or not jh.get_config('env.notifications.enabled'):
        return

    try:
        response = requests.get(
            f'https://api.telegram.org/bot{token}/sendMessage?chat_id={chat_id}&parse_mode=Markdown&text={msg}'
        )
        if response.status_code // 100 != 2:
            err_msg = f'Telegram ERROR [{response.status_code}]: {response.text}'
            if response.status_code // 100 == 4:
                err_msg += f'\nParameters: {msg}'
            logger.error(err_msg, send_notification=False)
    except requests.exceptions.ConnectionError:
        logger.error('Telegram ERROR: ConnectionError', send_notification=False)


def _discord(msg: str, webhook_address=None) -> None:
    from jesse.services import logger

    if not jh.get_config('env.notifications.enabled'):
        return

    try:
        response = requests.post(webhook_address, {'content': msg})
        if response.status_code // 100 != 2:
            err_msg = f'Discord ERROR [{response.status_code}]: {response.text}'
            if response.status_code // 100 == 4:
                err_msg += f'\nParameters: {msg}'
            logger.error(err_msg, send_notification=False)
    except requests.exceptions.ConnectionError:
        logger.error('Discord ERROR: ConnectionError', send_notification=False)


def _slack(msg: str, webhook_address) -> None:
    from jesse.services import logger

    if not jh.get_config('env.notifications.enabled'):
        return

    payload = {
        "text": msg
    }

    try:
        response = requests.post(webhook_address, json=payload)
        if response.status_code // 100 != 2:
            err_msg = f'Slack ERROR [{response.status_code}]: {response.text}'
            if response.status_code // 100 == 4:
                err_msg += f'\nParameters: {msg}'
            logger.error(err_msg, send_notification=False)
    except requests.exceptions.ConnectionError:
        logger.error('Slack ERROR: ConnectionError', send_notification=False)


def _custom_channel_notification(msg: dict):
    webhook = msg['webhook']

    if webhook.startswith('https://hooks.slack.com'):
        # a slack webhook
        _slack(msg['content'], webhook)
    elif webhook.startswith('https://discord.com/api/webhooks'):
        # a discord webhook
        _discord(msg['content'], webhook)
    else:
        raise ValueError(f'Custom Webhook {webhook}. seems to be neither a discord or slack webhook')


def _format_msg(msg: str) -> str:
    # if "_" exists in the message, replace it with "\_"
    msg = msg.replace('_', '\_')
    # # if "*" exists in the message, replace it with "\*"
    # msg = msg.replace('*', '\*')
    # # if "[" exists in the message, replace it with "\["
    # msg = msg.replace('[', '\[')
    # # if "]" exists in the message, replace it with "\}"
    # msg = msg.replace(']', '\]')
    return msg



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

# Import the high-performance Rust implementation
from jesse_rust import stochf as stochf_rust  # type: ignore

StochasticFast = namedtuple('StochasticFast', ['k', 'd'])

def stochf(candles: np.ndarray, fastk_period: int = 5, fastd_period: int = 3, fastd_matype: int = 0,
           sequential: bool = False) -> StochasticFast:
    """
    Stochastic Fast

    :param candles: np.ndarray
    :param fastk_period: int - default: 5
    :param fastd_period: int - default: 3
    :param fastd_matype: int - default: 0
    :param sequential: bool - default: False

    :return: StochasticFast(k, d)
    """
    if fastd_matype == 24 or fastd_matype == 29:
        raise ValueError("VWMA (matype 24) and VWAP (matype 29) cannot be used in stochf indicator.")

    candles = slice_candles(candles, sequential)

    # Convert to float64 for Rust compatibility
    candles_f64 = np.asarray(candles, dtype=np.float64)
    
    # Call the Rust implementation
    k, d = stochf_rust(candles_f64, fastk_period, fastd_period, fastd_matype)
    
    if sequential:
        return StochasticFast(k, d)
    else:
        return StochasticFast(k[-1], d[-1])



from typing import Union
import numpy as np
from jesse.helpers import get_candle_source, slice_candles


def vwap(
        candles: np.ndarray, source_type: str = "hlc3", anchor: str = "D", sequential: bool = False
) -> Union[float, np.ndarray]:
    """
    VWAP

    :param candles: np.ndarray
    :param source_type: str - default: "hlc3"
    :param anchor: str - default: "D"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    # Check if jesse_rust is available
    try:
        import jesse_rust
        
        candles = slice_candles(candles, sequential)
        
        # Use the Rust implementation with anchoring support
        result = jesse_rust.vwap(candles, source_type, anchor, sequential)
        
        if sequential:
            return result
        else:
            return None if np.isnan(result[-1]) else result[-1]
    
    except ImportError:
        # Fallback to original Python implementation with anchoring
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)
        
        # Convert timestamps to period indices
        timestamps = candles[:, 0].astype('datetime64[ms]').astype(f'datetime64[{anchor}]')
        group_indices = np.zeros(len(timestamps), dtype=np.int64)
        
        # Mark the start of each new period
        group_indices[1:] = (timestamps[1:] != timestamps[:-1]).astype(np.int64)
        group_indices = np.cumsum(group_indices)
        
        vwap_values = _calculate_vwap(source, candles[:, 5], group_indices)

        if sequential:
            return vwap_values
        else:
            return None if np.isnan(vwap_values[-1]) else vwap_values[-1]


def _calculate_vwap(source: np.ndarray, volume: np.ndarray, group_indices: np.ndarray) -> np.ndarray:
    """
    Calculate VWAP values with anchoring logic (pure Python implementation fallback)
    """
    vwap_values = np.zeros_like(source)
    cum_vol = 0.0
    cum_vol_price = 0.0
    current_group = group_indices[0]
    
    for i in range(len(source)):
        if group_indices[i] != current_group:
            cum_vol = 0.0
            cum_vol_price = 0.0
            current_group = group_indices[i]
            
        vol_price = volume[i] * source[i]
        cum_vol_price += vol_price
        cum_vol += volume[i]
        
        vwap_values[i] = cum_vol_price / cum_vol if cum_vol != 0 else np.nan
        
    return vwap_values



from collections import namedtuple

import numpy as np
from .vwma import vwma

from jesse.helpers import slice_candles

VWMACD = namedtuple('VWMACD', ['macd', 'signal', 'hist'])


def vwmacd(candles: np.ndarray, fast_period: int = 12, slow_period: int = 26, signal_period: int = 9,
           sequential: bool = False) -> VWMACD:
    """
    @author David.
    credits: https://www.tradingview.com/script/33Y1LzRq-Volume-Weighted-Moving-Average-Convergence-Divergence-MACD/
    VWMACD - Volume Weighted Moving Average Convergence/Divergence

    :param candles: np.ndarray
    :param fast_period: int - default: 12
    :param slow_period: int - default: 26
    :param signal_period: int - default: 9
    :param sequential: bool - default: False

    :return: VWMACD(macd, signal, hist)
    """
    fastWMA = vwma(candles, fast_period, sequential=True)
    slowWMA = vwma(candles, slow_period, sequential=True)

    macd_val = fastWMA - slowWMA
    signal = vwma(macd_val, signal_period, sequential=True)
    hist = macd_val - signal

    if sequential:
        return VWMACD(macd_val, signal, hist)
    else:
        return VWMACD(macd_val[-1], signal[-1], hist[-1])



import requests
import jesse.helpers as jh
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange
from typing import Union
from .hyperliquid_utils import timeframe_to_interval


class HyperliquidPerpetualMain(CandleExchange):
    def __init__(self, name: str, rest_endpoint: str) -> None:
        from jesse.modes.import_candles_mode.drivers.Binance.BinanceSpot import BinanceSpot

        super().__init__(name=name, count=5000, rate_limit_per_second=10, backup_exchange_class=BinanceSpot)
        self.name = name
        self.endpoint = rest_endpoint

    def get_starting_time(self, symbol: str) -> int:
        base_symbol = jh.get_base_asset(symbol)
        payload = {
            'type': 'candleSnapshot',
            'req': {
                'coin': base_symbol,
                'interval': 'W',
                'startTime': 1514811660
            }
        }
        headers = {
            'Content-Type': 'application/json',
        }

        response = requests.post(self.endpoint, json=payload, headers=headers)
        data = response.json()
        # Reverse the data list
        data = data[::-1]

        return int(data[1]['t'])

    def fetch(self, symbol: str, start_timestamp: int, timeframe: str = '1m') -> Union[list, None]:
        base_symbol = jh.get_base_asset(symbol)
        interval = timeframe_to_interval(timeframe)
        payload = {
            'type': 'candleSnapshot',
            'req': {
                'coin': base_symbol,
                'interval': interval,
                'startTime': int(start_timestamp)
            }
        }

        headers = {
            'Content-Type': 'application/json',
        }
        response = requests.post(self.endpoint, json=payload, headers=headers)
        self.validate_response(response)

        data = response.json()

        return [
            {
                'id': jh.generate_unique_id(),
                'exchange': self.name,
                'symbol': symbol,
                'timeframe': timeframe,
                'timestamp': int(d['t']),
                'open': float(d['o']),
                'close': float(d['c']),
                'high': float(d['h']),
                'low': float(d['l']),
                'volume': float(d['v'])
            } for d in data
        ]

    def get_available_symbols(self) -> list:
        response = requests.post(self.endpoint, json={'type': 'meta'})
        self.validate_response(response)
        data = response.json()['universe']
        pairs = []
        for item in data:
            pairs.append(item['name'] + '-USD')

        return list(sorted(pairs))



from typing import Optional
from fastapi import APIRouter, Header
from starlette.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.multiprocessing import process_manager
from jesse.services.web import ImportCandlesRequestJson, CancelRequestJson, GetCandlesRequestJson, DeleteCandlesRequestJson
import jesse.helpers as jh

router = APIRouter(prefix="/candles", tags=["Candles"])


@router.post("/import")
def import_candles(request_json: ImportCandlesRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Import candles for a specific exchange and symbol
    """
    jh.validate_cwd()

    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes import import_candles_mode

    process_manager.add_task(
        import_candles_mode.run, 
        request_json.id, 
        request_json.exchange, 
        request_json.symbol,
        request_json.start_date
    )

    return JSONResponse({'message': 'Started importing candles...'}, status_code=202)


@router.post("/cancel-import")
def cancel_import_candles(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    """
    Cancel an import candles process
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.cancel_process(request_json.id)

    return JSONResponse({'message': f'Candles process with ID of {request_json.id} was requested for termination'},
                        status_code=202)


@router.post("/clear-cache")
def clear_candles_database_cache(authorization: Optional[str] = Header(None)):
    """
    Clear the candles database cache
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.cache import cache
    cache.flush()

    return JSONResponse({
        'status': 'success',
        'message': 'Candles database cache cleared successfully',
    }, status_code=200)


@router.post("/get")
def get_candles(json_request: GetCandlesRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get candles for a specific exchange, symbol, and timeframe
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    jh.validate_cwd()

    from jesse.modes.data_provider import get_candles as gc

    arr = gc(json_request.exchange, json_request.symbol, json_request.timeframe)

    return JSONResponse({
        'id': json_request.id,
        'data': arr
    }, status_code=200)


@router.post("/existing")
def get_existing_candles(authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Get all existing candles in the database
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.candle import get_existing_candles
    
    try:
        data = get_existing_candles()
        return JSONResponse({'data': data}, status_code=200)
    except Exception as e:
        return JSONResponse({'error': str(e)}, status_code=500)


@router.post("/delete")
def delete_candles(json_request: DeleteCandlesRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    """
    Delete candles for a specific exchange and symbol
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.candle import delete_candles
    
    try:
        delete_candles(json_request.exchange, json_request.symbol)
        return JSONResponse({'message': 'Candles deleted successfully'}, status_code=200)
    except Exception as e:
        return JSONResponse({'error': str(e)}, status_code=500)



from collections import namedtuple
import numpy as np
from jesse_rust import ichimoku_cloud as ichimoku_cloud_rust

IchimokuCloud = namedtuple('IchimokuCloud', ['conversion_line', 'base_line', 'span_a', 'span_b'])


def ichimoku_cloud(candles: np.ndarray, conversion_line_period: int = 9, base_line_period: int = 26,
                   lagging_line_period: int = 52, displacement: int = 26) -> IchimokuCloud:
    """
    Ichimoku Cloud

    :param candles: np.ndarray
    :param conversion_line_period: int - default: 9
    :param base_line_period: int - default: 26
    :param lagging_line_period: int - default: 52
    :param displacement: - default: 26

    :return: IchimokuCloud(conversion_line, base_line, span_a, span_b)
    """
    if candles.shape[0] < 80:
        return IchimokuCloud(np.nan, np.nan, np.nan, np.nan)

    if candles.shape[0] > 80:
        candles = candles[-80:]
        
    conversion_line, base_line, span_a, span_b = ichimoku_cloud_rust(
        candles, conversion_line_period, base_line_period, 
        lagging_line_period, displacement
    )
    
    return IchimokuCloud(conversion_line, base_line, span_a, span_b)



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles

from .supersmoother import supersmoother_fast


def trendflex(candles: np.ndarray, period: int = 20, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Trendflex indicator by John F. Ehlers

    :param candles: np.ndarray
    :param period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)


    ssf = supersmoother_fast(source, period / 2)

    tf = trendflex_fast(ssf, period)

    if sequential:
        return tf
    else:
        return None if np.isnan(tf[-1]) else tf[-1]


@njit(cache=True)
def trendflex_fast(ssf, period):
    tf = np.full_like(ssf, 0)
    ms = np.full_like(ssf, 0)
    sums = np.full_like(ssf, 0)

    for i in range(ssf.shape[0]):
        if i >= period:
            my_sum = 0
            for t in range(1, period + 1):
                my_sum = my_sum + ssf[i] - ssf[i - t]
            my_sum /= period
            sums[i] = my_sum

            ms[i] = 0.04 * sums[i] * sums[i] + 0.96 * ms[i - 1]
            if ms[i] != 0:
                tf[i] = sums[i] / np.sqrt(ms[i])
    return tf



from collections import namedtuple
import numpy as np
from numba import njit
from jesse.helpers import same_length, slice_candles

FisherTransform = namedtuple('FisherTransform', ['fisher', 'signal'])


@njit
def _fisher_transform(high: np.ndarray, low: np.ndarray, period: int) -> tuple:
    """
    Numba-optimized implementation of Fisher Transform
    """
    length = len(high)
    mid_price = (high + low) / 2
    fisher = np.zeros(length)
    fisher_signal = np.zeros(length)

    # Initialize first value
    value1 = 0.0

    for i in range(period, length):
        # Find the highest high and lowest low in the period
        max_h = np.max(mid_price[i-period+1:i+1])
        min_l = np.min(mid_price[i-period+1:i+1])

        # Avoid division by zero
        if max_h - min_l == 0:
            value0 = 0
        else:
            value0 = 0.33 * 2 * ((mid_price[i] - min_l) / (max_h - min_l) - 0.5) + 0.67 * value1

        if value0 > 0.99:
            value0 = 0.999
        elif value0 < -0.99:
            value0 = -0.999

        fisher[i] = 0.5 * np.log((1 + value0) / (1 - value0)) + 0.5 * fisher[i-1]
        fisher_signal[i] = fisher[i-1]
        value1 = value0

    return fisher, fisher_signal


def fisher(candles: np.ndarray, period: int = 9, sequential: bool = False) -> FisherTransform:
    """
    The Fisher Transform helps identify price reversals.

    :param candles: np.ndarray
    :param period: int - default: 9
    :param sequential: bool - default: False

    :return: FisherTransform(fisher, signal)
    """
    candles = slice_candles(candles, sequential)

    fisher_val, fisher_signal = _fisher_transform(
        np.ascontiguousarray(candles[:, 3]),
        np.ascontiguousarray(candles[:, 4]),
        period
    )

    if sequential:
        return FisherTransform(same_length(candles, fisher_val), same_length(candles, fisher_signal))
    else:
        return FisherTransform(fisher_val[-1], fisher_signal[-1])



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def trange(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    TRANGE - True Range

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]

    true_range = np.empty_like(high)
    true_range[0] = high[0] - low[0]

    if len(candles) > 1:
        diff_hl = high[1:] - low[1:]
        diff_hpc = np.abs(high[1:] - close[:-1])
        diff_lpc = np.abs(low[1:] - close[:-1])
        true_range[1:] = np.maximum(np.maximum(diff_hl, diff_hpc), diff_lpc)

    res = true_range
    return res if sequential else res[-1]



import os
import requests
import jesse.helpers as jh
from jesse.info import exchange_info, jesse_supported_timeframes, JESSE_API_URL


def get_general_info(has_live=False) -> dict:
    from jesse.version import __version__ as jesse_version
    system_info = {
        'jesse_version': jesse_version
    }
    plan_info = {'plan': 'guest'}
    limits = {}

    if has_live:
        from jesse.services.auth import get_access_token
        access_token = get_access_token()
        if not access_token:
            has_live = False

        # version info
        from jesse_live.version import __version__ as live_version
        system_info['live_plugin_version'] = live_version

        if access_token:
            # get the account plan info via the access_token
            response = requests.post(
                JESSE_API_URL + '/v2/user-info',
                headers={'Authorization': f'Bearer {access_token}'}
            )
            if response.status_code != 200:
                raise Exception(
                    f"{response.status_code} error: {response.json()['message']}"
                )
            plan_info = response.json()
            limits = plan_info['limits']

    strategies_path = os.getcwd() + "/strategies/"
    strategies = list(sorted([name for name in os.listdir(strategies_path) if os.path.isdir(strategies_path + name) and not name.startswith('.')]))
    if "__pycache__" in strategies:
        strategies.remove("__pycache__")

    system_info['python_version'] = '{}.{}'.format(*jh.python_version())
    system_info['operating_system'] = jh.get_os()
    system_info['cpu_cores'] = jh.cpu_cores_count()
    system_info['is_docker'] = jh.is_docker()

    update_info = {}

    try:
        response = requests.get('https://pypi.org/pypi/jesse/json')
        update_info['jesse_latest_version'] = response.json()['info']['version']
        response = requests.get(
            JESSE_API_URL + '/plugins/live/releases/info'
        )
        update_info['jesse_live_latest_version'] = response.json()[0]['version']
        update_info['is_update_info_available'] = True
    except Exception:
        update_info['is_update_info_available'] = False

    res = {
        'exchanges': exchange_info,
        'strategies': strategies,
        'jesse_supported_timeframes': jesse_supported_timeframes,
        'has_live_plugin_installed': has_live,
        'system_info': system_info,
        'update_info': update_info,
        'plan': plan_info['plan'],
    }

    if has_live:
        res['limits'] = {
            'ip_limit': limits['ip_limit'],
            'live_trading_tabs': limits['live_trading_tabs'],
            'trading_routes': limits['trading_routes'],
            'data_routes': limits['data_routes'],
            'timeframes': limits['timeframes'],
            'exchanges': list(limits['exchanges'].keys()),
        }

    return res



import numpy as np
from numba import njit

import jesse.helpers as jh
import jesse.services.logger as logger
from jesse.enums import sides, order_types
from jesse.exceptions import InsufficientMargin
from jesse.models import Order
from jesse.services import selectors
from jesse.models.Exchange import Exchange


class FuturesExchange(Exchange):
    def __init__(
            self,
            name: str,
            starting_balance: float,
            fee_rate: float,
            futures_leverage_mode: str,
            futures_leverage: int
    ):
        super().__init__(name, starting_balance, fee_rate, 'futures')

        # # # # live-trading only # # # #
        # in futures trading, margin is only with one asset, so:
        self._available_margin = 0
        # in futures trading, wallet is only with one asset, so:
        self._wallet_balance = 0
        # so is started_balance
        self._started_balance = 0
        # # # # # # # # # # # # # # # # #

        self.futures_leverage_mode = futures_leverage_mode
        self.futures_leverage = futures_leverage

    @property
    def started_balance(self) -> float:
        if jh.is_livetrading():
            return self._started_balance

        return self.starting_assets[jh.app_currency()]

    @property
    def wallet_balance(self) -> float:
        if jh.is_livetrading():
            return self._wallet_balance

        return self.assets[self.settlement_currency]

    @property
    def available_margin(self) -> float:
        if jh.is_livetrading():
            return self._available_margin

        # In both live trading and backtesting/paper trading, we start with the balance
        margin = self.wallet_balance

        # Calculate the total spent amount considering leverage
        # Here we need to calculate the total cost of all open positions and orders, considering leverage
        total_spent = 0
        for asset in self.assets:
            if asset == self.settlement_currency:
                continue

            position = selectors.get_position(self.name, f"{asset}-{self.settlement_currency}")
            if position and position.is_open:
                # Adding the cost of open positions
                total_spent += position.total_cost
                # add unrealized PNL
                total_spent -= position.pnl

            # Summing up the cost of open orders (buy and sell), considering leverage
            sum_buy_orders = (self.buy_orders[asset][:][:, 0] * self.buy_orders[asset][:][:, 1]).sum()
            sum_sell_orders = (self.sell_orders[asset][:][:, 0] * self.sell_orders[asset][:][:, 1]).sum()

            total_spent += max(
                abs(sum_buy_orders) / self.futures_leverage, abs(sum_sell_orders) / self.futures_leverage
            )

        # Subtracting the total spent from the margin
        margin -= total_spent

        return margin

    def charge_fee(self, amount: float) -> None:
        if jh.is_livetrading():
            return

        fee_amount = abs(amount) * self.fee_rate
        new_balance = self.assets[self.settlement_currency] - fee_amount
        if fee_amount != 0:
            logger.info(
                f'Charged {round(fee_amount, 2)} as fee. Balance for {self.settlement_currency} on {self.name} changed from {round(self.assets[self.settlement_currency], 2)} to {round(new_balance, 2)}'
            )
        self.assets[self.settlement_currency] = new_balance

    def add_realized_pnl(self, realized_pnl: float) -> None:
        if jh.is_livetrading():
            return

        new_balance = self.assets[self.settlement_currency] + realized_pnl
        logger.info(
            f'Added realized PNL of {round(realized_pnl, 2)}. Balance for {self.settlement_currency} on {self.name} changed from {round(self.assets[self.settlement_currency], 2)} to {round(new_balance, 2)}')
        self.assets[self.settlement_currency] = new_balance

    def on_order_submission(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        base_asset = jh.base_asset(order.symbol)

        # make sure we don't spend more than we're allowed considering current allowed leverage
        if not order.reduce_only:
            # Calculate the effective order size considering leverage
            effective_order_size = abs(order.qty * order.price) / self.futures_leverage

            if effective_order_size > self.available_margin:
                raise InsufficientMargin(
                    f'Cannot submit an order with a value of ${round(order.qty * order.price)} when your available margin is ${round(self.available_margin)}. Consider increasing leverage number from the settings or reducing the order size.'
                )

        self.available_assets[base_asset] += order.qty

        if not order.reduce_only:
            if order.side == sides.BUY:
                self.buy_orders[base_asset].append(np.array([order.qty, order.price]))
            else:
                self.sell_orders[base_asset].append(np.array([order.qty, order.price]))

    def on_order_execution(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        base_asset = jh.base_asset(order.symbol)

        if not order.reduce_only:
            order_array = np.array([order.qty, order.price])
            if order.side == sides.BUY:
                item_index = np.where(np.all(self.buy_orders[base_asset].array == order_array, axis=1))[0]
                if len(item_index) > 0:
                    index = item_index[0]
                    self.buy_orders[base_asset].delete(index, axis=0)
            else:
                item_index = np.where(np.all(self.sell_orders[base_asset].array == order_array, axis=1))[0]
                if len(item_index) > 0:
                    index = item_index[0]
                    self.sell_orders[base_asset].delete(index, axis=0)

    def on_order_cancellation(self, order: Order) -> None:
        if jh.is_livetrading():
            return

        base_asset = jh.base_asset(order.symbol)

        self.available_assets[base_asset] -= order.qty
        if not order.reduce_only:
            order_array = np.array([order.qty, order.price])
            if order.side == sides.BUY:
                index = find_order_index(self.buy_orders[base_asset].array, order_array)
                if index != -1:
                    self.buy_orders[base_asset].delete(index, axis=0)
            else:
                index = find_order_index(self.sell_orders[base_asset].array, order_array)
                if index != -1:
                    self.sell_orders[base_asset].delete(index, axis=0)

    def update_from_stream(self, data: dict) -> None:
        """
        Used for updating the exchange from the WS stream (only for live trading)
        """
        if not jh.is_livetrading():
            raise Exception('This method is only for live trading')
        self._available_margin = data['available_margin']
        self._wallet_balance = data['wallet_balance']
        if self._started_balance == 0:
            self._started_balance = self._wallet_balance


@njit(cache=True)
def find_order_index(orders, order_array):
    for i in range(len(orders)):
        if np.all(orders[i] == order_array):
            return i
    return -1



from jesse.strategies import Strategy


# test_on_route_increased_position_and_on_route_reduced_position_and_strategy_vars part 2 - ETH-USD
class Test30(Strategy):
    def update_position(self):
        # increase position size
        if self.price == 20:
            self.buy = 1, self.price

        elif self.price == 50:
            self.take_profit = 1, self.price

        # close position with take_profit
        if self.price == 70:
            self.take_profit = self.position.qty, self.price

    def should_long(self):
        return self.price == 10

    def should_short(self):
        return False

    def go_long(self):
        self.buy = 1, self.price

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



from typing import Union
import numpy as np
from jesse.helpers import slice_candles
from jesse_rust import adx as adx_rust


def adx(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    ADX - Average Directional Movement Index

    :param candles: np.ndarray, expected 2D array with OHLCV data where index 3 is high, index 4 is low, and index 2 is close
    :param period: int - default: 14
    :param sequential: bool - if True, return full series, else return last value
    :return: float | np.ndarray
    """
    if len(candles.shape) < 2:
        raise ValueError("adx indicator requires a 2D array of candles")
        
    candles = slice_candles(candles, sequential)

    # Ensure there's enough data
    if len(candles) <= 2 * period:
        return np.nan if not sequential else np.full(len(candles), np.nan)

    result = adx_rust(candles, period)

    return result if sequential else result[-1]



from typing import Union
import numpy as np
from jesse.helpers import slice_candles
from jesse_rust import atr as rust_atr


def atr(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    ATR - Average True Range using optimized Rust implementation

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    result = rust_atr(candles, period)
    
    return result if sequential else result[-1]



from typing import List, Dict
import copy


def backtest(
        config: dict,
        routes: List[Dict[str, str]],
        data_routes: List[Dict[str, str]],
        candles: dict,
        warmup_candles: dict = None,
        generate_tradingview: bool = False,
        generate_hyperparameters: bool = False,
        generate_equity_curve: bool = False,
        benchmark: bool = False,
        generate_csv: bool = False,
        generate_json: bool = False,
        generate_logs: bool = False,
        hyperparameters: dict = None,
        fast_mode: bool = False
) -> dict:
    """
    An isolated backtest() function which is perfect for using in research, and AI training
    such as our own optimization mode. Because of it being a pure function, it can be used
    in Python's multiprocessing without worrying about pickling issues.

    Example `config`:
    {
        'starting_balance': 5_000,
        'fee': 0.005,
        'type': 'futures',
        'futures_leverage': 3,
        'futures_leverage_mode': 'cross',
        'exchange': 'Binance',
        'warm_up_candles': 0
    }

    Example `route`:
    [{'exchange': 'Bybit USDT Perpetual', 'strategy': 'A1', 'symbol': 'BTC-USDT', 'timeframe': '1m'}]

    Example `data_route`:
    [{'exchange': 'Bybit USDT Perpetual', 'symbol': 'BTC-USDT', 'timeframe': '3m'}]

    Example `candles`:
    {
        'Binance-BTC-USDT': {
            'exchange': 'Binance',
            'symbol': 'BTC-USDT',
            'candles': np.array([]),
        },
    }
    """
    return _isolated_backtest(
        config,
        routes,
        data_routes,
        candles,
        warmup_candles,
        run_silently=True,
        hyperparameters=hyperparameters,
        generate_tradingview=generate_tradingview,
        generate_csv=generate_csv,
        generate_json=generate_json,
        generate_equity_curve=generate_equity_curve,
        benchmark=benchmark,
        generate_hyperparameters=generate_hyperparameters,
        generate_logs=generate_logs,
        fast_mode=fast_mode,
    )


def _isolated_backtest(
        config: dict,
        routes: List[Dict[str, str]],
        data_routes: List[Dict[str, str]],
        candles: dict,
        warmup_candles: dict = None,
        run_silently: bool = True,
        hyperparameters: dict = None,
        generate_tradingview: bool = False,
        generate_csv: bool = False,
        generate_json: bool = False,
        generate_equity_curve: bool = False,
        benchmark: bool = False,
        generate_hyperparameters: bool = False,
        generate_logs: bool = False,
        fast_mode: bool = False,
) -> dict:
    from jesse.services.validators import validate_routes
    from jesse.modes.backtest_mode import simulator
    from jesse.config import config as jesse_config, reset_config
    from jesse.routes import router
    from jesse.store import store
    from jesse.config import set_config
    from jesse.services.candle import inject_warmup_candles_to_store
    import jesse.helpers as jh

    jesse_config['app']['trading_mode'] = 'backtest'

    # inject (formatted) configuration values
    set_config(_format_config(config))

    # set routes
    router.initiate(routes, data_routes)

    validate_routes(router)

    # initiate candle store
    store.candles.init_storage(5000)

    # assert that the passed candles are 1m candles
    for key, value in candles.items():
        candle_set = value['candles']
        if candle_set[1][0] - candle_set[0][0] != 60_000:
            raise ValueError(
                f'Candles passed to the research.backtest() must be 1m candles. '
                f'\nIf you wish to trade other timeframes, notice that you need to pass it through '
                f'the timeframe option in your routes. '
                f'\nThe difference between your candles are {candle_set[1][0] - candle_set[0][0]} milliseconds which more than '
                f'the accepted 60000 milliseconds.'
            )

    # make a copy to make sure we don't mutate the past data causing some issues for multiprocessing tasks
    trading_candles_dict = copy.deepcopy(candles)
    warmup_candles_dict = copy.deepcopy(warmup_candles)

    # if warmup_candles is passed, use it
    if warmup_candles:
        for c in jesse_config['app']['considering_candles']:
            key = jh.key(c[0], c[1])
            # inject warm-up candles
            inject_warmup_candles_to_store(
                warmup_candles_dict[key]['candles'],
                c[0],
                c[1]
            )

    # run backtest simulation
    backtest_result = simulator(
        trading_candles_dict,
        run_silently,
        hyperparameters=hyperparameters,
        generate_tradingview=generate_tradingview,
        generate_csv=generate_csv,
        generate_json=generate_json,
        generate_equity_curve=generate_equity_curve,
        benchmark=benchmark,
        generate_hyperparameters=generate_hyperparameters,
        generate_logs=generate_logs,
        fast_mode=fast_mode,
    )

    result = {
        'metrics': {'total': 0, 'win_rate': 0, 'net_profit_percentage': 0},
        'logs': None,
    }

    if backtest_result['metrics'] is None:
        result['metrics'] = {'total': 0, 'win_rate': 0, 'net_profit_percentage': 0}
    else:
        result['metrics'] = backtest_result['metrics']

    if generate_tradingview:
        result['tradingview'] = backtest_result['tradingview']
    if generate_csv:
        result['csv'] = backtest_result['csv']
    if generate_json:
        result['json'] = backtest_result['json']
    if generate_equity_curve:
        result['equity_curve'] = backtest_result['equity_curve']
    if generate_hyperparameters:
        result['hyperparameters'] = backtest_result['hyperparameters']
    if generate_logs:
        result['logs'] = backtest_result['logs']

    # reset store and config so rerunning would be flawlessly possible
    reset_config()
    store.reset()

    return result


def _format_config(config):
    """
    Jesse's required format for user_config is different from what this function accepts (so it
    would be easier to write for the researcher). Hence, we need to reformat the config_dict:
    """
    exchange_config = {
        'balance': config['starting_balance'],
        'fee': config['fee'],
        'type': config['type'],
        'name': config['exchange'],
    }
    # futures exchange has different config, so:
    if exchange_config['type'] == 'futures':
        exchange_config['futures_leverage'] = config['futures_leverage']
        exchange_config['futures_leverage_mode'] = config['futures_leverage_mode']

    return {
        'exchanges': {
            config['exchange']: exchange_config
        },
        'logging': {
            'balance_update': True,
            'order_cancellation': True,
            'order_execution': True,
            'order_submission': True,
            'position_closed': True,
            'position_increased': True,
            'position_opened': True,
            'position_reduced': True,
            'shorter_period_candles': False,
            'trading_candles': True
        },
        'warm_up_candles': config['warm_up_candles']
    }



from typing import Union

import numpy as np

from jesse.helpers import same_length, slice_candles


def marketfi(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    MARKETFI - Market Facilitation Index
    Formula: (High - Low) / Volume

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    # high is at index 3, low at index 4, volume at index 5
    res = (candles[:, 3] - candles[:, 4]) / candles[:, 5]

    return same_length(candles, res) if sequential else res[-1]



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse.indicators.ma import ma
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad

# Try to import the high-performance Rust implementation
try:
    from jesse_rust import bollinger_bands as bb_rust  # type: ignore
    from jesse_rust import moving_std
except ImportError:  # pragma: no cover
    bb_rust = None  # type: ignore
    from jesse_rust import moving_std


BollingerBands = namedtuple('BollingerBands', ['upperband', 'middleband', 'lowerband'])


def _bollinger_bands_fallback(source: np.ndarray, period: int, devup: float, devdn: float, matype: int, devtype: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Fallback implementation for non-SMA cases or when Rust is unavailable."""
    if devtype == 0:
        dev = moving_std(source, period)
    elif devtype == 1:
        dev = mean_ad(source, period, sequential=True)
    elif devtype == 2:
        dev = median_ad(source, period, sequential=True)
    else:
        raise ValueError("devtype not in (0, 1, 2)")

    middlebands = ma(source, period=period, matype=matype, sequential=True)
    upperbands = middlebands + devup * dev
    lowerbands = middlebands - devdn * dev
    
    return upperbands, middlebands, lowerbands


def bollinger_bands(
        candles: np.ndarray,
        period: int = 20,
        devup: float = 2,
        devdn: float = 2,
        matype: int = 0,
        devtype: int = 0,
        source_type: str = "close",
        sequential: bool = False
) -> BollingerBands:
    """
    BBANDS - Bollinger Bands

    :param candles: np.ndarray
    :param period: int - default: 20
    :param devup: float - default: 2
    :param devdn: float - default: 2
    :param matype: int - default: 0
    :param devtype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: BollingerBands(upperband, middleband, lowerband)
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    # Use optimized Rust implementation for standard case (SMA + standard deviation)
    if bb_rust is not None and matype == 0 and devtype == 0:
        upperbands, middlebands, lowerbands = bb_rust(source.astype(np.float64), period, devup, devdn)
    else:
        # Handle special cases or fallback
        if matype == 24 or matype == 29:
            # VWMA or VWAP need the full candles array
            upperbands, middlebands, lowerbands = _bollinger_bands_fallback(candles, period, devup, devdn, matype, devtype)
        else:
            upperbands, middlebands, lowerbands = _bollinger_bands_fallback(source, period, devup, devdn, matype, devtype)

    if sequential:
        return BollingerBands(upperbands, middlebands, lowerbands)
    else:
        return BollingerBands(upperbands[-1], middlebands[-1], lowerbands[-1])



def import_candles(
    exchange: str,
    symbol: str,
    start_date: str,
    show_progressbar: bool = True,
) -> str:
    from jesse.modes.import_candles_mode import run

    return run(
        client_id='',
        exchange=exchange,
        symbol=symbol,
        start_date_str=start_date,
        running_via_dashboard=False,
        show_progressbar=show_progressbar
    )



from typing import Union

import numpy as np

from jesse.helpers import slice_candles


def obv(candles: np.ndarray, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    OBV - On Balance Volume

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    close = candles[:, 2]
    volume = candles[:, 5]
    # Compute the change in OBV: add volume if price increases, subtract if decreases, else 0
    delta = np.where(close[1:] > close[:-1], volume[1:], np.where(close[1:] < close[:-1], -volume[1:], 0))
    obv_arr = np.empty_like(volume, dtype=np.float64)
    obv_arr[0] = volume[0]
    if len(volume) > 1:
        obv_arr[1:] = volume[0] + np.cumsum(delta)
    return obv_arr if sequential else obv_arr[-1]



from typing import Union

import numpy as np

from jesse.helpers import slice_candles
from jesse_rust import willr as willr_rust


def willr(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    WILLR - Williams' %R

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    
    # Convert to float64 for Rust compatibility
    candles_f64 = np.asarray(candles, dtype=np.float64)
    
    # Call the Rust implementation
    result = willr_rust(candles_f64, period)
    
    return result if sequential else result[-1]



from collections import namedtuple

from jesse.indicators.sma import sma
import numpy as np

from jesse.helpers import slice_candles

VPCI = namedtuple('VPCI', ['vpci', 'vpcis'])


def vpci(candles: np.ndarray, short_range: int = 5, long_range: int = 25, sequential: bool = False) -> VPCI:
    """
    VPCI - Volume Price Confirmation Indicator

    :param candles: np.ndarray
    :param short_range: int - default: 5
    :param long_range: int - default: 25
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    vwma_long = sma(candles[:, 2] * candles[:, 5], long_range, sequential=True) / sma(candles[:, 5], long_range, sequential=True)
    VPC = vwma_long - sma(candles[:, 2], long_range, sequential=True)

    vwma_short = sma(candles[:, 2] * candles[:, 5], short_range, sequential=True) / sma(candles[:, 5], short_range, sequential=True)
    VPR = vwma_short / sma(candles[:, 2], short_range, sequential=True)

    VM = sma(candles[:, 5], short_range, sequential=True) / sma(candles[:, 5], long_range, sequential=True)
    VPCI_val = VPC * VPR * VM

    VPCIS = sma(VPCI_val * candles[:, 5], short_range, sequential=True) / sma(candles[:, 5], short_range, sequential=True)

    if sequential:
        return VPCI(VPCI_val, VPCIS)
    else:
        return VPCI(VPCI_val[-1], VPCIS[-1])



from typing import Union

import numpy as np
from numba import njit

import jesse.helpers as jh
from jesse.helpers import slice_candles
import jesse_rust as jr


@njit(cache=True)
def _ema(arr: np.ndarray, period: int) -> np.ndarray:
    """
    Compute the exponential moving average (EMA) using a simple for loop, accelerated with numba.
    The formula is: EMA[i] = sum_{j=0}^{i} (alpha * (1-alpha)**(i - j) * arr[j]) where alpha = 2/(period+1).
    This is computed iteratively:
      EMA[0] = alpha * arr[0]
      EMA[i] = alpha * arr[i] + (1 - alpha) * EMA[i-1] for i >= 1
    """
    alpha = 2.0 / (period + 1)
    n = arr.shape[0]
    result = np.empty(n, dtype=arr.dtype)
    result[0] = alpha * arr[0]
    for i in range(1, n):
        result[i] = alpha * arr[i] + (1 - alpha) * result[i - 1]
    return result


def dti(candles: np.ndarray, r: int = 14, s: int = 10, u: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    DTI by William Blau calculated using Rust implementation for better performance.

    :param candles: np.ndarray of candles data
    :param r: period for the first EMA smoothing (default 14)
    :param s: period for the second EMA smoothing (default 10)
    :param u: period for the third EMA smoothing (default 5)
    :param sequential: if True, returns the full sequence, otherwise only the last value
    :return: float or np.ndarray of DTI values
    """
    candles = slice_candles(candles, sequential)
    
    # Use the Rust implementation
    res = jr.dti(candles, r, s, u)
    
    if sequential:
        return res
    else:
        return None if np.isnan(res[-1]) else res[-1]



import numpy as np

from typing import Union

from jesse.helpers import get_candle_source, slice_candles


def rocr(candles: np.ndarray, period: int = 10, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    ROCR - Rate of change ratio: (price / price_lagged)

    :param candles: np.ndarray
    :param period: int - default: 10
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    # Initialize an array of NaNs with the same shape as source
    res = np.full(source.shape, np.nan, dtype=float)
    
    # Only compute if there are enough data points
    if source.shape[0] > period:
        # Compute rate of change ratio using vectorized operation; for indices < period, remains NaN
        res[period:] = source[period:] / source[:-period]

    return res if sequential else res[-1]



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


def cwma(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Cubed Weighted Moving Average

    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = vpwma_fast(source, period)

    return res if sequential else res[-1]


@njit(cache=True)
def vpwma_fast(source, period):
    newseries = np.copy(source)
    for j in range(period + 1, source.shape[0]):
        my_sum = 0.0
        weightSum = 0.0
        for i in range(period - 1):
            weight = np.power(period - i, 3)
            my_sum += (source[j - i] * weight)
            weightSum += weight
        newseries[j] = my_sum / weightSum
    return newseries



from collections import namedtuple

import numpy as np
from jesse_rust import donchian as rust_donchian
from jesse.helpers import slice_candles

DonchianChannel = namedtuple('DonchianChannel', ['upperband', 'middleband', 'lowerband'])


def donchian(candles: np.ndarray, period: int = 20, sequential: bool = False) -> DonchianChannel:
    """
    Donchian Channels

    :param candles: np.ndarray
    :param period: int - default: 20
    :param sequential: bool - default: False

    :return: DonchianChannel(upperband, middleband, lowerband)
    """
    candles = slice_candles(candles, sequential)
    
    if sequential:
        # Use optimized Rust implementation
        rust_result = rust_donchian(candles, period)
        return DonchianChannel(
            rust_result['upperband'],
            rust_result['middleband'],
            rust_result['lowerband']
        )
    else:
        # Non-sequential mode only needs the very last value of the channel.
        # Instead of calling the Rust implementation (which processes the
        # entire candle history and incurs extra FFI overhead), we can obtain
        # the same result in pure NumPy by looking at just the last `period`
        # candles.

        if candles.shape[0] < period:
            # Not enough candles yet → behave exactly like the Rust variant
            # which would return NaNs for the incomplete window.
            return DonchianChannel(np.nan, np.nan, np.nan)

        # Slice only the window we need.
        window = candles[-period:]

        # Candle columns: 0 -> timestamp, 1 -> open, 2 -> close, 3 -> high, 4 -> low
        highs = window[:, 3]
        lows = window[:, 4]

        upper = highs.max()
        lower = lows.min()
        middle = (upper + lower) / 2

        return DonchianChannel(upper, middle, lower)



from fastapi import APIRouter, WebSocket, WebSocketDisconnect, Query
import asyncio
import json
from asyncio import Queue

from jesse.services import auth as authenticator
from jesse.services.redis import async_redis
from jesse.services.multiprocessing import process_manager
import jesse.helpers as jh

router = APIRouter(tags=["WebSocket"])


@router.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket, token: str = Query(...)):
    from jesse.services.env import ENV_VALUES

    if not authenticator.is_valid_token(token):
        return

    await websocket.accept()

    queue = Queue()
    ch, = await async_redis.psubscribe(f"{ENV_VALUES['APP_PORT']}:channel:*")

    async def echo(q):
        try:
            while True:
                msg = await q.get()
                msg = json.loads(msg)
                msg['id'] = process_manager.get_client_id(msg['id'])
                await websocket.send_json(msg)
        except WebSocketDisconnect:
            await async_redis.punsubscribe(f"{ENV_VALUES['APP_PORT']}:channel:*")
            print(jh.color('WebSocket disconnected', 'yellow'))
        except Exception as e:
            print(jh.color(str(e), 'red'))

    async def reader(channel, q):
        async for ch, message in channel.iter():
            await q.put(message)

    asyncio.get_running_loop().create_task(reader(ch, queue))
    asyncio.get_running_loop().create_task(echo(queue))

    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        await async_redis.punsubscribe(f"{ENV_VALUES['APP_PORT']}:channel:*")
        print(jh.color('WebSocket disconnected', 'yellow'))



import pytest
from jesse import research


def test_store_candles():
    """
    for now, don't actually store it in the db. But test validations, etc
    """
    # validate that candles type must be np.ndarray
    with pytest.raises(TypeError):
        research.store_candles({}, 'Test Exchange', 'BTC-USDT')
    with pytest.raises(TypeError):
        research.store_candles([], 'Test Exchange', 'BTC-USDT')

    # add validation for timeframe to make sure it's `1m`
    with pytest.raises(ValueError):
        close_prices = [10, 11]
        np_candles = research.candles_from_close_prices(close_prices)
        # change the timeframe to 5 minutes
        np_candles[1][0] += 300_000
        research.store_candles(np_candles, 'Test Exchange', 'BTC-USDT')

    # just make sure it doesn't raise an error
    close_prices = [10, 11, 12, 12, 11, 13, 14, 12, 11, 15]
    np_candles = research.candles_from_close_prices(close_prices)
    research.store_candles(np_candles, 'Test Exchange', 'BTC-USDT')



from jesse.strategies import Strategy


class TestDnaMethod(Strategy):
    def before(self):
        if self.index == 0:
            assert self.hp == {'profit_target': 7, 'qty_w': 86}

    def hyperparameters(self):
        return [
            {'name': 'qty_w', 'type': int, 'min': 10, 'max': 95, 'default': 70},
            {'name': 'profit_target', 'type': int, 'min': 1, 'max': 40, 'default': 5},
        ]

    def dna(self):
        return "o4"

    def should_long(self) -> bool:
        return False

    def should_short(self) -> bool:
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False



import jesse.helpers as jh
import jesse.services.selectors as selectors
from jesse.config import reset_config
from jesse.enums import timeframes, exchanges
from jesse.factories import range_candles
from jesse.modes import backtest_mode
from jesse.routes import router
from jesse.store import store
from jesse.config import config


def test_backtesting_one_route():
    reset_config()
    routes = [
        {'symbol': 'BTC-USDT', 'timeframe': timeframes.MINUTE_5, 'strategy': 'Test19'}
    ]
    config['env']['exchanges'][exchanges.SANDBOX]['type'] = 'futures'

    candles = {}
    key = jh.key(exchanges.SANDBOX, 'BTC-USDT')
    candles[key] = {
        'exchange': exchanges.SANDBOX,
        'symbol': 'BTC-USDT',
        'candles': range_candles(5 * 20)
    }

    # run backtest (dates are fake just to pass)
    backtest_mode.run('000', False, {}, exchanges.SANDBOX, routes, [], '2019-04-01', '2019-04-02', candles)

    one_min = store.candles.get_candles(exchanges.SANDBOX, 'BTC-USDT', '1m')
    five_min = store.candles.get_candles(exchanges.SANDBOX, 'BTC-USDT', '5m')

    # assert the count of present candles
    assert len(five_min) == 20
    assert len(one_min) == 20 * 5

    first_1 = one_min[0]
    last_1 = one_min[-1]
    first_5 = five_min[0]
    last_5 = five_min[-1]

    # assert time in store
    assert store.app.time == last_1[0] + 60000

    # assert timestamps
    assert first_1[0] == first_5[0]
    assert last_1[0] == (last_5[0] + 60000 * 4)

    # there must be only one positions present
    assert len(store.positions.storage) == 1
    p = selectors.get_position(exchanges.SANDBOX, 'BTC-USDT')
    assert p.is_close
    assert p.current_price == last_1[2]
    assert p.current_price == last_5[2]

    # assert routes
    assert len(router.routes) == 1
    assert router.routes[0].exchange == exchanges.SANDBOX
    assert router.routes[0].symbol == 'BTC-USDT'
    assert router.routes[0].timeframe == '5m'
    assert router.routes[0].strategy_name == 'Test19'
    # assert that the strategy has been initiated
    assert router.routes[0].strategy is not None


def test_backtesting_three_routes():
    reset_config()
    routes = [
        {'symbol': 'BTC-USDT', 'timeframe': timeframes.MINUTE_5, 'strategy': 'Test19'},
        {'symbol': 'ETH-USDT', 'timeframe': timeframes.MINUTE_5, 'strategy': 'Test19'},
        {'symbol': 'XRP-USDT', 'timeframe': timeframes.MINUTE_15, 'strategy': 'Test19'}
    ]
    config['env']['exchanges'][exchanges.SANDBOX]['type'] = 'futures'

    candles = {}
    for r in routes:
        key = jh.key(exchanges.SANDBOX, r['symbol'])
        candles[key] = {
            'exchange': exchanges.SANDBOX,
            'symbol': r['symbol'],
            'candles': range_candles(5 * 3 * 20)
        }

    # run backtest (dates are fake just to pass)
    backtest_mode.run('000', False, {}, exchanges.SANDBOX, routes, [], '2019-04-01', '2019-04-02', candles)

    # there must be three positions present with the updated current_price
    assert len(store.positions.storage) == 3

    for r in router.routes:
        # r3's '15m' timeframe makes r1 and r2 to support
        # '15' timeframe as well. r1 and r2 also make r3
        # to support '5m' timeframe also.
        r_one_min = store.candles.get_candles(r.exchange, r.symbol, '1m')
        r_five_min = store.candles.get_candles(r.exchange, r.symbol, '5m')
        r_fifteen_min = store.candles.get_candles(r.exchange, r.symbol, '15m')

        # assert the count of present candles
        assert len(r_one_min) == (5 * 3) * 20
        assert len(r_five_min) == 20 * 3
        assert len(r_fifteen_min) == 20

        r_first_1 = r_one_min[0]
        r_last_1 = r_one_min[-1]
        r_first_5 = r_five_min[0]
        r_last_5 = r_five_min[-1]
        r_last_15 = r_fifteen_min[-1]

        # assert timestamps
        assert r_first_1[0] == r_first_5[0]
        assert r_last_1[0] == (r_last_5[0] + 60000 * 4)
        assert r_last_5[0] == (r_last_15[0] + 60000 * 10)

        # assert positions
        p = selectors.get_position(r.exchange, r.symbol)
        assert p.is_close is True
        last_candle = store.candles.get_candles(r.exchange, r.symbol, '1m')[-1]
        assert p.current_price == last_candle[2]

        # assert that the strategy has been initiated
        assert r.strategy is not None



from typing import Union
from jesse.indicators.sma import sma
import numpy as np


def vosc(candles: np.ndarray, short_period: int = 2, long_period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    VOSC - Volume Oscillator

    :param candles: np.ndarray
    :param short_period: int - default: 2
    :param long_period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    from jesse.helpers import same_length, slice_candles

    candles = slice_candles(candles, sequential)
    volume = candles[:, 5]

    short_sma = sma(volume, short_period, sequential=True)
    long_sma = sma(volume, long_period, sequential=True)

    vosc_values = (short_sma - long_sma) / long_sma * 100

    vosc_result = same_length(candles, vosc_values)

    return vosc_result if sequential else vosc_result[-1]



from typing import Union
import numpy as np
import jesse.helpers as jh
from jesse_rust import cvi as cvi_rust


def cvi(candles: np.ndarray, period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    CVI - Chaikins Volatility

    :param candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = jh.slice_candles(candles, sequential)
    
    # Call the Rust implementation
    res = cvi_rust(candles, period)

    return jh.same_length(candles, res) if sequential else res[-1]



from jesse import exceptions
import jesse.helpers as jh
from jesse.services import logger


def validate_routes(router) -> None:
    if not router.routes:
        raise exceptions.RouteNotFound(
            'No routes found. Please add at least one route at: routes.py\nMore info: https://docs.jesse.trade/docs/routes.html#routing')

    # validation for number of routes in the live mode
    if jh.is_live():
        if len(router.routes) > 10:
            logger.broadcast_error_without_logging('Too many routes (not critical, but use at your own risk): Using that more than 5 routes in live/paper trading is not recommended because exchange WS connections are often not reliable for handling that much traffic.')



from abc import ABC, abstractmethod
from typing import Union
from jesse.models import Order

class Exchange(ABC):
    """
    The interface that every Exchange driver has to implement
    """

    @abstractmethod
    def market_order(self, symbol: str, qty: float, current_price: float, side: str, reduce_only: bool) -> Order:
        pass

    @abstractmethod
    def limit_order(self, symbol: str, qty: float, price: float, side: str, reduce_only: bool) -> Order:
        pass

    @abstractmethod
    def stop_order(self, symbol: str, qty: float, price: float, side: str, reduce_only: bool) -> Order:
        pass

    @abstractmethod
    def cancel_all_orders(self, symbol: str) -> None:
        pass

    @abstractmethod
    def cancel_order(self, symbol: str, order_id: str) -> None:
        pass

    @abstractmethod
    def _fetch_precisions(self) -> None:
        pass



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles
from jesse_rust import di as di_rust

DI = namedtuple('DI', ['plus', 'minus'])


def di(candles: np.ndarray, period: int = 14, sequential: bool = False) -> DI:
    """
    DI - Directional Indicator

    :param candles: np.ndarray, where columns are expected to be: index 2: close, index 3: high, index 4: low.
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: DI(plus, minus)
    """
    candles = slice_candles(candles, sequential)
    n = len(candles)
    if n < 2:
        if sequential:
            return DI(np.full(n, np.nan), np.full(n, np.nan))
        else:
            return DI(np.nan, np.nan)

    # Use Rust implementation
    plus_DI_arr, minus_DI_arr = di_rust(candles, period)
    if sequential:
        return DI(plus_DI_arr, minus_DI_arr)
    else:
        return DI(plus_DI_arr[-1], minus_DI_arr[-1])



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles
from jesse_rust import ema as ema_rust


def ema(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    EMA - Exponential Moving Average using Numba for optimization

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    result = ema_rust(source, period)
    return result if sequential else result[-1]



import numpy as np

from jesse.strategies import Strategy


class TestMetrics1(Strategy):
    def should_long(self) -> bool:
        return self.price == 10

    def should_short(self) -> bool:
        return False

    def go_long(self):
        self.buy = 10, 10
        # sell it for $50 profit
        self.take_profit = 10, 15

    def go_short(self):
        pass

    def should_cancel_entry(self):
        return False

    def on_close_position(self, order):
        assert self.metrics['total'] == 1
        assert self.metrics['starting_balance'] == 10000
        assert self.metrics['finishing_balance'] == 10050
        assert self.metrics['win_rate'] == 1
        assert self.metrics['ratio_avg_win_loss'] is np.nan
        assert self.metrics['longs_count'] == 1
        assert self.metrics['shorts_count'] == 0
        assert self.metrics['longs_percentage'] == 100
        assert self.metrics['shorts_percentage'] == 0
        assert self.metrics['fee'] == 0
        assert self.metrics['net_profit'] == 50
        assert self.metrics['net_profit_percentage'] == 0.5
        assert self.metrics['average_win'] == 50
        assert self.metrics['average_loss'] is np.nan
        assert self.metrics['expectancy'] == 50
        assert self.metrics['expectancy_percentage'] == 0.5
        assert self.metrics['expected_net_profit_every_100_trades'] == 50
        assert self.metrics['average_holding_period'] == 300
        assert self.metrics['average_losing_holding_period'] is np.nan
        assert self.metrics['average_winning_holding_period'] == 300
        assert self.metrics['gross_loss'] == 0
        assert self.metrics['gross_profit'] == 50
        assert self.metrics['open_pl'] == 0



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles
from .sma import sma

AO = namedtuple('AO', ['osc', 'change'])

def momentum(arr):
    ret = np.full(arr.shape, np.nan)
    if len(arr) > 1:
        ret[1:] = arr[1:] - arr[:-1]
    return ret

def ao(candles: np.ndarray, sequential: bool = False) -> AO:
    """
    Awesome Oscillator

    :param candles: np.ndarray
    :param sequential: bool - default: False

    :return: AO(osc, change)
    """
    candles = slice_candles(candles, sequential)

    # Calculate hl2 as (high+low)/2
    hl2 = (candles[:, 3] + candles[:, 4]) / 2
    # Calculate simple moving averages on hl2 for periods 5 and 34
    sma5 = sma(hl2, 5, sequential=True)
    sma34 = sma(hl2, 34, sequential=True)
    ao = sma5 - sma34

    # Calculate momentum as the difference between consecutive values
    mom = momentum(ao)

    if sequential:
        return AO(ao, mom)
    else:
        return AO(ao[-1], mom[-1])



from jesse.strategies import Strategy


class TestCurrentRouteIndex2(Strategy):
    def before(self) -> None:
        if self.index == 0 or self.index == 10:
            assert self.current_route_index == 1

    def should_long(self):
        return False

    def go_long(self):
        pass

    def should_cancel_entry(self):
        return False



import time
import re
from typing import Dict, List, Tuple
import numpy as np
import jesse.helpers as jh
import jesse.services.metrics as stats
import jesse.services.selectors as selectors
from jesse import exceptions
from jesse.config import config
from jesse.enums import timeframes, order_types
from jesse.models import Order, Position
from jesse.modes.utils import save_daily_portfolio_balance
from jesse.routes import router
from jesse.services import charts
from jesse.services import report
from jesse.services.candle import generate_candle_from_one_minutes, print_candle, candle_includes_price, split_candle, \
    get_candles, inject_warmup_candles_to_store
from jesse.services.file import store_logs
from jesse.services.validators import validate_routes
from jesse.store import store
from jesse.services import logger
from jesse.services.failure import register_custom_exception_handler
from jesse.services.redis import sync_publish, is_process_active
from timeloop import Timeloop
from datetime import timedelta
from jesse.services.progressbar import Progressbar
from jesse.constants import TIMEFRAME_TO_ONE_MINUTES


def run(
        client_id: str,
        debug_mode: bool,
        user_config: dict,
        exchange: str,
        routes: List[Dict[str, str]],
        data_routes: List[Dict[str, str]],
        start_date: str,
        finish_date: str,
        candles: dict = None,
        chart: bool = False,
        tradingview: bool = False,
        csv: bool = False,
        json: bool = False,
        fast_mode: bool = False,
        benchmark: bool = False
) -> None:
    if not jh.is_unit_testing():
        # at every second, we check to see if it's time to execute stuff
        status_checker = Timeloop()

        @status_checker.job(interval=timedelta(seconds=1))
        def handle_time():
            if is_process_active(client_id) is False:
                raise exceptions.Termination

        status_checker.start()

    from jesse.config import config
    config['app']['trading_mode'] = 'backtest'

    # debug flag
    config['app']['debug_mode'] = debug_mode

    register_custom_exception_handler()

    _execute_backtest(
        client_id, debug_mode, user_config, exchange, routes, data_routes, start_date, finish_date, candles, chart,
        tradingview, csv, json, fast_mode, benchmark
    )


def _execute_backtest(
        client_id: str,
        debug_mode: bool,
        user_config: dict,
        exchange: str,
        routes: List[Dict[str, str]],
        data_routes: List[Dict[str, str]],
        start_date: str,
        finish_date: str,
        candles: dict = None,
        chart: bool = False,
        tradingview: bool = False,
        csv: bool = False,
        json: bool = False,
        fast_mode: bool = False,
        benchmark: bool = False
):
    """
    Executes the backtest that has been initiated from within the dashboard. The purpose of extracting these
    functionalities into this function is so that in case it fails due to a missing data route, it can add
    it and then re-execute itself.
    """
    from jesse.config import set_config

    # inject config
    if not jh.is_unit_testing():
        set_config(user_config)
    # add exchange to routes
    for r in routes:
        r['exchange'] = exchange
    for r in data_routes:
        r['exchange'] = exchange
    # set routes
    router.initiate(routes, data_routes)

    store.app.set_session_id(client_id)

    # validate routes
    validate_routes(router)

    # initiate candle store
    store.candles.init_storage(5000)

    # load historical candles
    if candles is None:
        try:
            warmup_candles, candles = load_candles(
                jh.date_to_timestamp(start_date),
                jh.date_to_timestamp(finish_date)
            )
            _handle_warmup_candles(warmup_candles, start_date)
        except exceptions.CandlesNotFound as e:
            # Extract symbol and exchange from error message
            match = re.search(r"for (.*?) on (.*?)$", str(e))
            if match:
                symbol, exchange = match.groups()
                raise exceptions.CandlesNotFound({
                    'message': str(e),
                    'symbol': symbol,
                    'exchange': exchange,
                    'start_date': start_date,
                    'type': 'missing_candles'
                })
            raise e

    if not jh.should_execute_silently():
        sync_publish('general_info', {
            'session_id': jh.get_session_id(),
            'debug_mode': str(config['app']['debug_mode']),
        })
        # candles info
        key = f"{config['app']['considering_candles'][0][0]}-{config['app']['considering_candles'][0][1]}"
        sync_publish('candles_info', stats.candles_info(candles[key]['candles']))
        # routes info
        sync_publish('routes_info', stats.routes(router.routes))

    # run backtest simulation
    result = None
    try:
        result = simulator(
            candles,
            run_silently=jh.should_execute_silently(),
            generate_tradingview=tradingview,
            generate_csv=csv,
            generate_json=json,
            generate_equity_curve=True,
            benchmark=benchmark,
            generate_hyperparameters=True,
            fast_mode=fast_mode,
        )
    except exceptions.RouteNotFound as e:
        # Extract exchange, symbol, and timeframe using regular expressions
        match = re.search(r"symbol='(.+?)', timeframe='(.+?)'", str(e))
        if match:
            symbol = match.group(1)
            timeframe = match.group(2)
            # Adjust data_routes to include the missing route
            data_routes.append({
                'exchange': exchange,
                'symbol': symbol,
                'timeframe': timeframe
            })
            # to prevent an issue with warmupcandles being None
            candles = None
            # notify the user about the missing data route and retry the backtest simulation
            sync_publish('notification', {
                'message': f'Missing data route for "{symbol}" with "{timeframe}" timeframe. Adding it and retrying...',
                'type': 'error'
            })
            # retry the backtest simulation
            _execute_backtest(
                client_id, debug_mode, user_config, exchange, routes, data_routes, start_date, finish_date, candles,
                chart, tradingview, csv, json, fast_mode, benchmark
            )
        else:
            raise e

    if result and not jh.should_execute_silently():
        sync_publish('alert', {
            'message': f"Successfully executed backtest simulation in: {result['execution_duration']} seconds",
            'type': 'success'
        })
        sync_publish('hyperparameters', result['hyperparameters'])
        sync_publish('metrics', result['metrics'])
        sync_publish('equity_curve', result['equity_curve'], compression=True)
        sync_publish('trades', result['trades'], compression=True)
        if chart:
            sync_publish('candles_chart', _get_formatted_candles_for_frontend(), compression=True)
            sync_publish('orders_chart', _get_formatted_orders_for_frontend(), compression=True)
            sync_publish('add_line_to_candle_chart', _get_add_line_to_candle_chart(), compression=True)
            sync_publish('add_extra_line_chart', _get_add_extra_line_chart(), compression=True)
            sync_publish('add_horizontal_line_to_candle_chart', _get_add_horizontal_line_to_candle_chart(), compression=True)
            sync_publish('add_horizontal_line_to_extra_chart', _get_add_horizontal_line_to_extra_chart(), compression=True)

    # close database connection
    from jesse.services.db import database
    database.close_connection()


def _get_formatted_candles_for_frontend():
    arr = []
    for r in router.routes:
        candles_arr = store.candles.get_candles(r.exchange, r.symbol, r.timeframe)
        # Find the index where the starting time actually begins.
        starting_index = 0
        for i, c in enumerate(candles_arr):
            if c[0] >= store.app.starting_time:
                starting_index = i
                break

        candles = [{
            'time': int(c[0]/1000),
            'open': c[1],
            'close': c[2],
            'high': c[3],
            'low': c[4],
            'volume': c[5]
        } for c in candles_arr[starting_index:]]
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'candles': candles
        })
    return arr


def _get_formatted_orders_for_frontend():
    arr = []
    for r in router.routes:
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'orders': r.strategy._executed_orders
        })
    return arr


def _get_add_line_to_candle_chart():
    arr = []
    for r in router.routes:
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'lines': r.strategy._add_line_to_candle_chart_values
        })
    return arr


def _get_add_extra_line_chart():
    arr = []
    for r in router.routes:
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'charts': r.strategy._add_extra_line_chart_values
        })
    return arr


def _get_add_horizontal_line_to_candle_chart():
    arr = []
    for r in router.routes:
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'lines': r.strategy._add_horizontal_line_to_candle_chart_values
        })
    return arr


def _get_add_horizontal_line_to_extra_chart():
    arr = []
    for r in router.routes:
        arr.append({
            'exchange': r.exchange,
            'symbol': r.symbol,
            'timeframe': r.timeframe,
            'lines': r.strategy._add_horizontal_line_to_extra_chart_values
        })
    return arr


def _handle_missing_candles(exchange: str, symbol: str, start_date: int, message: str = None):
    """Helper function to handle missing candles scenarios"""
    formatted_date = jh.timestamp_to_date(start_date)
    if message is None:
        message = f'Missing trading candles for {symbol} on {exchange} from {formatted_date}'
    
    sync_publish(
        "missing_candles",
        {
            "message": message,
            "symbol": symbol,
            "exchange": exchange,
            "start_date": formatted_date,
        },
    )
    
    raise exceptions.CandlesNotFound({
        'message': message,
        'symbol': symbol,
        'exchange': exchange,
        'start_date': start_date,
        'type': 'missing_candles'
    })


def load_candles(start_date: int, finish_date: int) -> Tuple[dict, dict]:
    warmup_num = jh.get_config('env.data.warmup_candles_num', 210)
    max_timeframe = jh.max_timeframe(config['app']['considering_timeframes'])

    # load and add required warm-up candles for backtest, and then Prepare trading candles
    trading_candles = {}
    warmup_candles = {}
    for c in config['app']['considering_candles']:
        exchange, symbol = c[0], c[1]
        warmup_candles_arr, trading_candle_arr = get_candles(
            exchange, symbol, max_timeframe, start_date, finish_date, warmup_num, caching=True, is_for_jesse=True
        )

        # Ensure that trading_candle_arr is not None or empty
        if trading_candle_arr is None or (isinstance(trading_candle_arr, np.ndarray) and trading_candle_arr.size == 0):
            _handle_missing_candles(
                exchange, 
                symbol, 
                start_date, 
                f"Missing trading candles for {symbol} on {exchange}"
            )

        # Check that the first trading candle covers the requested start date.
        if trading_candle_arr[0][0] > start_date:
            _handle_missing_candles(exchange, symbol, start_date)

        # Check that the last trading candle covers the requested finish date.
        if trading_candle_arr[-1][0] < (finish_date - 60_000):
            _handle_missing_candles(exchange, symbol, start_date)

        # add trading candles
        trading_candles[jh.key(exchange, symbol)] = {
            'exchange': exchange,
            'symbol': symbol,
            'candles': trading_candle_arr
        }

        warmup_candles[jh.key(exchange, symbol)] = {
            'exchange': exchange,
            'symbol': symbol,
            'candles': warmup_candles_arr
        }

    return warmup_candles, trading_candles


def _handle_warmup_candles(warmup_candles: dict, start_date: str) -> None:
    try:
        for c in config['app']['considering_candles']:
            exchange, symbol = c[0], c[1]
            inject_warmup_candles_to_store(warmup_candles[jh.key(exchange, symbol)]['candles'], exchange, symbol)
    except ValueError as e:
        # Extract exchange and symbol from error message
        match = re.search(r"for (.*?)/(.*?)\?", str(e))
        if match:
            exchange, symbol = match.groups()
            
            # Calculate warmup start date using the same logic as load_candles()
            warmup_num = jh.get_config('env.data.warmup_candles_num', 210)
            max_timeframe = jh.max_timeframe(config['app']['considering_timeframes'])
            # Convert max_timeframe to minutes and multiply by warmup_num
            warmup_minutes = TIMEFRAME_TO_ONE_MINUTES[max_timeframe] * warmup_num
            warmup_start_timestamp = jh.date_to_timestamp(start_date) - (warmup_minutes * 60_000)
            warmup_start_date = jh.timestamp_to_date(warmup_start_timestamp)
            # Publish the missing candles error to the frontend
            # This will trigger the alert in the BacktestTab.vue component
            # so that the user can import the missing candles
            sync_publish(
                "missing_candles",
                {
                    "message": f'Missing warmup candles for {symbol} on {exchange} from {warmup_start_date}',
                    "symbol": symbol,
                    "exchange": exchange,
                    "start_date": warmup_start_date,
                },
            )
            raise exceptions.CandlesNotFound(str(e))
        raise e


def simulator(*args, fast_mode: bool = False, **kwargs) -> dict:
    if fast_mode:
        return _skip_simulator(*args, **kwargs)

    return _step_simulator(*args, **kwargs)


def _step_simulator(
        candles: dict,
        run_silently: bool,
        hyperparameters: dict = None,
        generate_tradingview: bool = False,
        generate_csv: bool = False,
        generate_json: bool = False,
        generate_equity_curve: bool = False,
        benchmark: bool = False,
        generate_hyperparameters: bool = False,
        generate_logs: bool = False,
) -> dict:
    # In case generating logs is specifically demanded, the debug mode must be enabled.
    if generate_logs:
        config['app']['debug_mode'] = True

    begin_time_track = time.time()

    key = f"{config['app']['considering_candles'][0][0]}-{config['app']['considering_candles'][0][1]}"
    first_candles_set = candles[key]['candles']

    length = _simulation_minutes_length(candles)
    _prepare_times_before_simulation(candles)
    _prepare_routes(hyperparameters)

    # add initial balance
    save_daily_portfolio_balance(is_initial=True)

    progressbar = Progressbar(length, step=420)
    last_update_time = None
    for i in range(length):
        # update time
        store.app.time = first_candles_set[i][0] + 60_000

        # add candles
        for j in candles:
            short_candle = candles[j]['candles'][i]
            if i != 0:
                previous_short_candle = candles[j]['candles'][i - 1]
                short_candle = _get_fixed_jumped_candle(previous_short_candle, short_candle)
            exchange = candles[j]['exchange']
            symbol = candles[j]['symbol']

            store.candles.add_candle(short_candle, exchange, symbol, '1m', with_execution=False,
                                     with_generation=False)

            # print short candle
            if jh.is_debuggable('shorter_period_candles'):
                print_candle(short_candle, True, symbol)

            _simulate_price_change_effect(short_candle, exchange, symbol)

            # generate and add candles for bigger timeframes
            for timeframe in config['app']['considering_timeframes']:
                # for 1m, no work is needed
                if timeframe == '1m':
                    continue

                count = TIMEFRAME_TO_ONE_MINUTES[timeframe]
                # until = count - ((i + 1) % count)

                if (i + 1) % count == 0:
                    generated_candle = generate_candle_from_one_minutes(
                        timeframe,
                        candles[j]['candles'][(i - (count - 1)):(i + 1)]
                    )

                    store.candles.add_candle(generated_candle, exchange, symbol, timeframe, with_execution=False,
                                             with_generation=False)

        last_update_time = _update_progress_bar(progressbar, run_silently, i, candle_step=420,
                                                last_update_time=last_update_time)

        # now that all new generated candles are ready, execute
        for r in router.routes:
            count = TIMEFRAME_TO_ONE_MINUTES[r.timeframe]
            # 1m timeframe
            if r.timeframe == timeframes.MINUTE_1:
                r.strategy._execute()
            elif (i + 1) % count == 0:
                # print candle
                if jh.is_debuggable('trading_candles'):
                    print_candle(store.candles.get_current_candle(r.exchange, r.symbol, r.timeframe), False,
                                 r.symbol)
                r.strategy._execute()

            store.orders.update_active_orders(r.exchange, r.symbol)

        # now check to see if there's any MARKET orders waiting to be executed
        _execute_market_orders()

        if i != 0 and i % 1440 == 0:
            save_daily_portfolio_balance()

    _finish_progress_bar(progressbar, run_silently)

    execution_duration = 0
    if not run_silently:
        # print executed time for the backtest session
        finish_time_track = time.time()
        execution_duration = round(finish_time_track - begin_time_track, 2)

    for r in router.routes:
        r.strategy._terminate()
        _execute_market_orders()

    # now that backtest simulation is finished, add finishing balance
    save_daily_portfolio_balance()

    # set the ending time for the backtest session
    store.app.ending_time = store.app.time + 60_000

    result = _generate_outputs(
        candles,
        generate_tradingview=generate_tradingview,
        generate_csv=generate_csv,
        generate_json=generate_json,
        generate_equity_curve=generate_equity_curve,
        benchmark=benchmark,
        generate_hyperparameters=generate_hyperparameters,
        generate_logs=generate_logs,
    )
    result['execution_duration'] = execution_duration
    return result


def _simulation_minutes_length(candles: dict) -> int:
    key = f"{config['app']['considering_candles'][0][0]}-{config['app']['considering_candles'][0][1]}"
    first_candles_set = candles[key]["candles"]
    return len(first_candles_set)


def _prepare_times_before_simulation(candles: dict) -> None:
    # result = {}
    # begin_time_track = time.time()
    key = f"{config['app']['considering_candles'][0][0]}-{config['app']['considering_candles'][0][1]}"
    first_candles_set = candles[key]["candles"]
    # length = len(first_candles_set)
    # to preset the array size for performance
    try:
        store.app.starting_time = first_candles_set[0][0]
    except IndexError:
        raise IndexError('Check your "warm_up_candles" config value')
    store.app.time = first_candles_set[0][0]


def _prepare_routes(hyperparameters: dict = None) -> None:
    # initiate strategies
    for r in router.routes:
        # if the r.strategy is str read it from file
        if isinstance(r.strategy_name, str):
            StrategyClass = jh.get_strategy_class(r.strategy_name)
        # else it is a class object so just use it
        else:
            StrategyClass = r.strategy_name

        try:
            r.strategy = StrategyClass()
        except TypeError:
            raise exceptions.InvalidStrategy(
                "Strategy validation failed. Make sure your strategy has the mandatory methods such as should_long(), "
                "go_long(), etc. For working examples, visit: https://jesse.trade/strategies"
            )
        except:
            raise

        r.strategy.name = r.strategy_name
        r.strategy.exchange = r.exchange
        r.strategy.symbol = r.symbol
        r.strategy.timeframe = r.timeframe

        # read the dna from strategy's dna() and use it for injecting inject hyperparameters
        # first convert DNS string into hyperparameters
        if len(r.strategy.dna()) > 0 and hyperparameters is None:
            hyperparameters = jh.dna_to_hp(
                r.strategy.hyperparameters(), r.strategy.dna()
            )

        # inject hyperparameters sent within the optimize mode
        if hyperparameters is not None:
            r.strategy.hp = hyperparameters

        # init few objects that couldn't be initiated in Strategy __init__
        # it also injects hyperparameters into self.hp in case the route does not uses any DNAs
        r.strategy._init_objects()

        selectors.get_position(r.exchange, r.symbol).strategy = r.strategy


def _update_progress_bar(
        progressbar: Progressbar, run_silently: bool, candle_index: int, candle_step: int, last_update_time: float
) -> float:
    throttle_interval = 0.5
    current_time = time.time()
    if not run_silently and candle_index % candle_step == 0:
        progressbar.update()

        if last_update_time is None or (current_time - last_update_time) >= throttle_interval:
            sync_publish(
                "progressbar",
                {
                    "current": progressbar.current,
                    "estimated_remaining_seconds": progressbar.estimated_remaining_seconds,
                },
            )
            # Update the last update time
            last_update_time = current_time

    # Return the last update time for future reference
    return last_update_time


def _finish_progress_bar(progressbar: Progressbar, run_silently: bool):
    if run_silently:
        return

    progressbar.finish()
    sync_publish(
        "progressbar",
        {
            "current": 100,
            "estimated_remaining_seconds": 0,
        },
    )


def _get_fixed_jumped_candle(
        previous_candle: np.ndarray, candle: np.ndarray
) -> np.ndarray:
    """
    A little workaround for the times that the price has jumped and the opening
    price of the current candle is not equal to the previous candle's close!

    :param previous_candle: np.ndarray
    :param candle: np.ndarray
    """
    if previous_candle[2] < candle[1]:
        candle[1] = previous_candle[2]
        candle[4] = min(previous_candle[2], candle[4])
    elif previous_candle[2] > candle[1]:
        candle[1] = previous_candle[2]
        candle[3] = max(previous_candle[2], candle[3])

    return candle


def _simulate_price_change_effect(real_candle: np.ndarray, exchange: str, symbol: str) -> None:
    current_temp_candle = real_candle.copy()
    executed_order = False

    executing_orders = _get_executing_orders(exchange, symbol, real_candle)
    if len(executing_orders) > 1:
        # extend the candle shape from (6,) to (1,6)
        executing_orders = _sort_execution_orders(executing_orders, current_temp_candle[None, :])

    while True:
        if len(executing_orders) == 0:
            executed_order = False
        else:
            for index, order in enumerate(executing_orders):
                if index == len(executing_orders) - 1 and not order.is_active:
                    executed_order = False

                if not order.is_active:
                    continue

                if candle_includes_price(current_temp_candle, order.price):
                    storable_temp_candle, current_temp_candle = split_candle(current_temp_candle, order.price)
                    _update_all_routes_a_partial_candle(exchange, symbol, storable_temp_candle)

                    p = selectors.get_position(exchange, symbol)
                    p.current_price = storable_temp_candle[2]

                    executed_order = True

                    order.execute()
                    executing_orders = _get_executing_orders(exchange, symbol, current_temp_candle)
                    if len(executing_orders) > 1:
                        # extend the candle shape from (6,) to (1,6)
                        executing_orders = _sort_execution_orders(executing_orders, current_temp_candle[None, :])

                    # break from the for loop, we'll try again inside the while
                    # loop with the new current_temp_candle
                    break
                else:
                    executed_order = False

        if not executed_order:
            # add/update the real_candle to the store so we can move on
            store.candles.add_candle(
                real_candle, exchange, symbol, '1m',
                with_execution=False,
                with_generation=False
            )
            p = selectors.get_position(exchange, symbol)
            if p:
                p.current_price = real_candle[2]
            break

    _check_for_liquidations(real_candle, exchange, symbol)


def _check_for_liquidations(candle: np.ndarray, exchange: str, symbol: str) -> None:
    p: Position = selectors.get_position(exchange, symbol)

    if not p:
        return

    # for now, we only support the isolated mode:
    if p.mode != 'isolated':
        return

    if candle_includes_price(candle, p.liquidation_price):
        closing_order_side = jh.closing_side(p.type)

        # create the market order that is used as the liquidation order
        order = Order({
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': exchange,
            'side': closing_order_side,
            'type': order_types.MARKET,
            'reduce_only': True,
            'qty': jh.prepare_qty(p.qty, closing_order_side),
            'price': p.bankruptcy_price
        })

        store.orders.add_order(order)

        store.app.total_liquidations += 1

        logger.info(f'{p.symbol} liquidated at {p.liquidation_price}')

        order.execute()


def _generate_outputs(
        candles: dict,
        generate_tradingview: bool = False,
        generate_csv: bool = False,
        generate_json: bool = False,
        generate_equity_curve: bool = False,
        benchmark: bool = False,
        generate_hyperparameters: bool = False,
        generate_logs: bool = False,
):
    result = {}
    if generate_hyperparameters:
        result["hyperparameters"] = stats.hyperparameters(router.routes)
    result["metrics"] = report.portfolio_metrics()
    result["trades"] = report.trades()
    # generate logs in json, csv and tradingview's pine-editor format
    logs_path = store_logs(generate_json, generate_tradingview, generate_csv)
    if generate_json:
        result["json"] = logs_path["json"]
    if generate_tradingview:
        result["tradingview"] = logs_path["tradingview"]
    if generate_csv:
        result["csv"] = logs_path["csv"]
    if generate_equity_curve:
        result["equity_curve"] = charts.equity_curve(benchmark)
    if generate_logs:
        result["logs"] = f"storage/logs/backtest-mode/{jh.get_session_id()}.txt"
    return result


def _skip_simulator(
        candles: dict,
        run_silently: bool,
        hyperparameters: dict = None,
        generate_tradingview: bool = False,
        generate_csv: bool = False,
        generate_json: bool = False,
        generate_equity_curve: bool = False,
        benchmark: bool = False,
        generate_hyperparameters: bool = False,
        generate_logs: bool = False,
) -> dict:
    # In case generating logs is specifically demanded, the debug mode must be enabled.
    if generate_logs:
        config["app"]["debug_mode"] = True

    begin_time_track = time.time()

    length = _simulation_minutes_length(candles)
    _prepare_times_before_simulation(candles)
    _prepare_routes(hyperparameters)

    # add initial balance
    save_daily_portfolio_balance(is_initial=True)

    candles_step = _calculate_minimum_candle_step()
    progressbar = Progressbar(length, step=candles_step)
    last_update_time = None
    for i in range(0, length, candles_step):
        # update time moved to _simulate_price_change_effect__multiple_candles
        # store.app.time = first_candles_set[i][0] + (60_000 * candles_step)
        _simulate_new_candles(candles, i, candles_step)

        last_update_time = _update_progress_bar(progressbar, run_silently, i, candles_step,
                                                last_update_time=last_update_time)

        _execute_routes(i, candles_step)

        # now check to see if there's any MARKET orders waiting to be executed
        _execute_market_orders()

        if i != 0 and i % 1440 == 0:
            save_daily_portfolio_balance()

    _finish_progress_bar(progressbar, run_silently)

    execution_duration = 0
    if not run_silently:
        # print executed time for the backtest session
        finish_time_track = time.time()
        execution_duration = round(finish_time_track - begin_time_track, 2)

    for r in router.routes:
        r.strategy._terminate()
        _execute_market_orders()

    # now that backtest simulation is finished, add finishing balance
    save_daily_portfolio_balance()

    # set the ending time for the backtest session
    store.app.ending_time = store.app.time + 60_000

    result = _generate_outputs(
        candles,
        generate_tradingview=generate_tradingview,
        generate_csv=generate_csv,
        generate_json=generate_json,
        generate_equity_curve=generate_equity_curve,
        benchmark=benchmark,
        generate_hyperparameters=generate_hyperparameters,
        generate_logs=generate_logs,
    )
    result['execution_duration'] = execution_duration
    return result


def _calculate_minimum_candle_step():
    """
    Calculates the minimum step for update candles that will allow simple updates on the simulator.
    """
    # config["app"]["considering_timeframes"] use '1m' also even if not required by the user so take only what the user
    # is requested.
    consider_time_frames = [
        TIMEFRAME_TO_ONE_MINUTES[route["timeframe"]]
        for route in router.all_formatted_routes
    ]
    return np.gcd.reduce(consider_time_frames)


def _simulate_new_candles(candles: dict, candle_index: int, candles_step: int) -> None:
    i = candle_index
    # add candles
    for j in candles:
        short_candles = candles[j]["candles"][i: i + candles_step]
        if i != 0:
            previous_short_candles = candles[j]["candles"][i - 1]
            # work the same, the fix needs to be done only on the gap of 1m edge candles.
            short_candles[0] = _get_fixed_jumped_candle(
                previous_short_candles, short_candles[0]
            )
        exchange = candles[j]["exchange"]
        symbol = candles[j]["symbol"]

        _simulate_price_change_effect_multiple_candles(
            short_candles, exchange, symbol
        )

        # generate and add candles for bigger timeframes
        for timeframe in config["app"]["considering_timeframes"]:
            # for 1m, no work is needed
            if timeframe == "1m":
                continue

            count = TIMEFRAME_TO_ONE_MINUTES[timeframe]

            if (i + candles_step) % count == 0:
                generated_candle = generate_candle_from_one_minutes(
                    timeframe,
                    candles[j]["candles"][
                    i - count + candles_step: i + candles_step],
                )

                store.candles.add_candle(
                    generated_candle,
                    exchange,
                    symbol,
                    timeframe,
                    with_execution=False,
                    with_generation=False,
                )


def _simulate_price_change_effect_multiple_candles(
        short_timeframes_candles: np.ndarray, exchange: str, symbol: str
) -> None:
    real_candle = np.array(
        [
            short_timeframes_candles[0][0],
            short_timeframes_candles[0][1],
            short_timeframes_candles[-1][2],
            short_timeframes_candles[:, 3].max(),
            short_timeframes_candles[:, 4].min(),
            short_timeframes_candles[:, 5].sum(),
        ]
    )
    executing_orders = _get_executing_orders(exchange, symbol, real_candle)
    if len(executing_orders) > 0:
        if len(executing_orders) > 1:
            executing_orders = _sort_execution_orders(executing_orders, short_timeframes_candles)

        for i in range(len(short_timeframes_candles)):
            current_temp_candle = short_timeframes_candles[i].copy()
            if i > 0:
                current_temp_candle[3] = max(current_temp_candle[3], short_timeframes_candles[i-1, 2])
                current_temp_candle[4] = min(current_temp_candle[4], short_timeframes_candles[i-1, 2])
            is_executed_order = False

            while True:
                if len(executing_orders) == 0:
                    is_executed_order = False
                else:
                    for index, order in enumerate(executing_orders):
                        if index == len(executing_orders) - 1 and not order.is_active:
                            is_executed_order = False
                        if not order.is_active:
                            continue

                        if candle_includes_price(current_temp_candle, order.price):
                            storable_temp_candle, current_temp_candle = split_candle(
                                current_temp_candle, order.price
                            )
                            _update_all_routes_a_partial_candle(
                                exchange,
                                symbol,
                                storable_temp_candle,
                            )
                            p = selectors.get_position(exchange, symbol)
                            p.current_price = storable_temp_candle[2]

                            is_executed_order = True

                            store.app.time = storable_temp_candle[0] + 60_000
                            order.execute()
                            executing_orders = _get_executing_orders(
                                exchange, symbol, real_candle
                            )

                            # break from the for loop, we'll try again inside the while
                            # loop with the new current_temp_candle
                            break
                        else:
                            is_executed_order = False

                if not is_executed_order:
                    # add/update the real_candle to the store so we can move on
                    store.candles.add_candle(
                        short_timeframes_candles[i].copy(),
                        exchange,
                        symbol,
                        "1m",
                        with_execution=False,
                        with_generation=False,
                    )
                    p = selectors.get_position(exchange, symbol)
                    if p:
                        p.current_price = current_temp_candle[2]
                    break

    store.candles.add_multiple_1m_candles(
        short_timeframes_candles,
        exchange,
        symbol,
    )
    store.app.time = real_candle[0] + (60_000 * len(short_timeframes_candles))
    _check_for_liquidations(real_candle, exchange, symbol)

    p = selectors.get_position(exchange, symbol)
    if p:
        p.current_price = short_timeframes_candles[-1, 2]


def _update_all_routes_a_partial_candle(
        exchange: str,
        symbol: str,
        storable_temp_candle: np.ndarray,
) -> None:
    """
    This function get called when an order is getting executed you need to update the other timeframe how their last
    candles looks like
    """
    store.candles.add_candle(
        storable_temp_candle,
        exchange,
        symbol,
        "1m",
        with_execution=False,
        with_generation=False,
    )

    for route in router.all_formatted_routes:
        timeframe = route['timeframe']
        if route['exchange'] != exchange or route['symbol'] != symbol:
            continue
        if timeframe == '1m':
            continue
        tf_minutes = TIMEFRAME_TO_ONE_MINUTES[timeframe]
        number_of_needed_candles = int(storable_temp_candle[0] % (tf_minutes * 60_000) // 60000) + 1
        candles_1m = store.candles.get_candles(exchange, symbol, '1m')[-number_of_needed_candles:]
        generated_candle = generate_candle_from_one_minutes(
            timeframe,
            candles_1m,
            accept_forming_candles=True
        )
        store.candles.add_candle(
            generated_candle,
            exchange,
            symbol,
            timeframe,
            with_execution=False,
            with_generation=False,
        )


def _execute_routes(candle_index: int, candles_step: int) -> None:
    # now that all new generated candles are ready, execute
    for r in router.routes:
        count = TIMEFRAME_TO_ONE_MINUTES[r.timeframe]
        # 1m timeframe
        if r.timeframe == timeframes.MINUTE_1:
            r.strategy._execute()
        elif (candle_index + candles_step) % count == 0:
            # print candle
            if jh.is_debuggable("trading_candles"):
                print_candle(
                    store.candles.get_current_candle(
                        r.exchange, r.symbol, r.timeframe
                    ),
                    False,
                    r.symbol,
                )
            r.strategy._execute()

        store.orders.update_active_orders(r.exchange, r.symbol)


def _execute_market_orders():
    store.orders.execute_pending_market_orders()


def _get_executing_orders(exchange, symbol, real_candle):
    orders = store.orders.get_active_orders(exchange, symbol)
    return [
        order
        for order in orders
        if order.is_active and candle_includes_price(real_candle, order.price)
    ]


def _sort_execution_orders(orders: List[Order], short_candles: np.ndarray):
    remaining_orders = set(orders)
    sorted_orders = []
    
    for candle in short_candles:
        open_price, close_price, low, high = candle[1], candle[2], candle[4], candle[3]

        # Did not use candle_includes_price() for performance, keeping it vectorization-friendly
        included_orders = [order for order in remaining_orders if low <= order.price <= high]

        if len(included_orders) == 1:
            sorted_orders.append(included_orders[0])
            remaining_orders.remove(included_orders[0])
        elif len(included_orders) > 1:
            # in case that the orders are above
            on_open, above_open, below_open = [], [], []
            for order in included_orders:
                if order.price == open_price:
                    on_open.append(order)
                if order.price > open_price:
                    above_open.append(order)
                else:
                    below_open.append(order)
            sorted_orders += on_open
            remaining_orders.difference_update(on_open)

            is_red = open_price > close_price
            if is_red:
                # heuristic that first the price goes up and then down, so this is the order execution sort
                above_open.sort(key=lambda o: o.price)
                below_open.sort(key=lambda o: o.price, reverse=True)
                sorted_orders += above_open + below_open
                remaining_orders.difference_update(above_open + below_open)
            else:
                below_open.sort(key=lambda o: o.price, reverse=True)
                above_open.sort(key=lambda o: o.price)
                sorted_orders += below_open + above_open
                remaining_orders.difference_update(below_open + above_open)

        if len(sorted_orders) == len(orders):
            break

    return sorted_orders
